from typing import Optional

import click

from anyscale.commands.util import OptionPromptNull
from anyscale.controllers.cloud_controller import CloudController


@click.group(
    "cloud",
    short_help="Configure cloud provider authentication for Anyscale.",
    help="""Configure cloud provider authentication and setup
to allow Anyscale to launch instances in your account.""",
)
def cloud_cli() -> None:
    pass


@cloud_cli.command(name="delete", help="Delete a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Delete cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to delete. Alternative to cloud name.",
    required=False,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
def cloud_delete(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str], yes: bool
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().delete_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id, skip_confirmation=yes
    )


@cloud_cli.command(
    name="set-default",
    help=(
        "Sets default cloud for your organization. This operation can only be performed "
        "by organization admins, and the default cloud must have organization level "
        "permissions."
    ),
)
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Set cloud as default by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to set as default. Alternative to cloud name.",
    required=False,
)
def cloud_set_default(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().set_default_cloud(
        cloud_name=cloud_name or name, cloud_id=cloud_id
    )


@cloud_cli.command(name="setup", help="Set up a cloud provider.")
@click.option(
    "--provider",
    help="The cloud provider type.",
    required=True,
    prompt="Provider",
    type=click.Choice(["aws", "gcp"], case_sensitive=False),
)
@click.option(
    "--region",
    cls=OptionPromptNull,
    help="Region to set up the credentials in.",
    required=True,
    prompt="Region",
    default_option="provider",
    default=lambda p: "us-west-2" if p == "aws" else "us-west1",
    show_default=True,
)
@click.option("--name", "-n", help="Name of the cloud.", required=True, prompt="Name")
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
@click.option(
    "--gcp-folder-id",
    help="Folder ID where the 'Anyscale'  folder will be created.",
    required=False,
    type=int,
)
@click.option(
    "--vpc-peering-ip-range", help="IP range for VPC peering.", required=False
)
@click.option(
    "--vpc-peering-target-project-id",
    help="Project ID of the target VPC.",
    required=False,
)
@click.option(
    "--vpc-peering-target-vpc-id", help="VPC ID of the target VPC.", required=False
)
def setup_cloud(
    provider: str,
    region: str,
    name: str,
    yes: bool,
    gcp_folder_id: Optional[int],
    vpc_peering_ip_range: Optional[str],
    vpc_peering_target_project_id: Optional[str],
    vpc_peering_target_vpc_id: Optional[str],
) -> None:
    CloudController().setup_cloud(
        provider=provider,
        region=region,
        name=name,
        yes=yes,
        folder_id=gcp_folder_id,
        vpc_peering_ip_range=vpc_peering_ip_range,
        vpc_peering_target_project_id=vpc_peering_target_project_id,
        vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
    )


@cloud_cli.command(
    name="list", help=("List information about clouds in your Anyscale organization."),
)
@click.option(
    "--name",
    "-n",
    required=False,
    default=None,
    help="Name of cloud to get information about.",
)
@click.option(
    "--cloud-id",
    "--id",
    required=False,
    default=None,
    help=("Id of cloud to get information about."),
)
def list_cloud(name: Optional[str], cloud_id: Optional[str],) -> None:
    print(CloudController().list_clouds(cloud_name=name, cloud_id=cloud_id,))


@cloud_cli.group("config", help="Manage the configuration for a cloud.")
def cloud_config_group() -> None:
    pass


@cloud_config_group.command("update", help="Update the configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to update. Alternative to cloud name.",
    required=False,
)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--max-stopped-instances",
    help="Maximum number of stopped instances permitted in the shared instance pool.",
    required=True,
)
def cloud_config_update(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    max_stopped_instances: int,
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().update_cloud_config(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        max_stopped_instances=max_stopped_instances,
    )


@cloud_config_group.command("get", help="Get the current configuration for a cloud.")
@click.argument("cloud-name", required=False)
@click.option("--name", "-n", help="Update configuration of cloud by name.", type=str)
@click.option(
    "--cloud-id",
    "--id",
    help="Cloud id to get details about. Alternative to cloud name.",
    required=False,
)
def cloud_config_get(
    cloud_name: Optional[str], name: Optional[str], cloud_id: Optional[str]
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    print(
        CloudController().get_cloud_config(
            cloud_name=cloud_name or name, cloud_id=cloud_id,
        )
    )


@cloud_cli.command(
    name="secrets",
    help="Allow clusters started in this cloud to access secrets manager.",
    hidden=True,
)
@click.argument("cloud-name", required=False)
@click.option(
    "--name",
    "-n",
    help="Allow clusters started in this cloud (identified by cloud name) to access secrets manager.",
    type=str,
)
@click.option(
    "--cloud-id",
    "--id",
    help="Allow clusters started in this cloud (identified by cloud id) to access secrets manager.",
    required=False,
)
@click.option(
    "--yes", "-y", is_flag=True, default=False, help="Don't ask for confirmation."
)
@click.option(
    "--write-permissions",
    is_flag=True,
    default=False,
    help="Allow clusters starting in this cloud to write to secrets manager as well.",
)
def cloud_secrets(
    cloud_name: Optional[str],
    name: Optional[str],
    cloud_id: Optional[str],
    write_permissions: bool,
    yes: bool,
) -> None:
    if cloud_name and name and cloud_name != name:
        raise click.ClickException(
            "The positional argument CLOUD_NAME and the keyword argument --name "
            "were both provided. Please only provide one of these two arguments."
        )
    CloudController().experimental_setup_secrets(
        cloud_name=cloud_name or name,
        cloud_id=cloud_id,
        write_permissions=write_permissions,
        yes=yes,
    )



import json
import logging
from typing import Any, Callable, Dict, Optional
from unittest.mock import ANY, Mock, patch

import pytest

from anyscale.connect_utils.start_interactive_session import (
    INITIAL_SCALE_TYPE,
    StartInteractiveSessionBlock,
)
from anyscale.utils.connect_helpers import (
    AnyscaleClientConnectResponse,
    AnyscaleClientContext,
)


def _connected(ray: Mock, ret: Dict[str, Any],) -> Callable[[Any, Any], Dict[str, Any]]:
    def connected(*a: Any, **kw: Any) -> Dict[str, Any]:
        ray.util.client.ray.is_connected.return_value = True
        returnable = {
            "num_clients": 1,
            "ray_version": "mock_ray_version",
            "ray_commit": "mock_ray_commit",
            "python_version": "mock_python_version",
            "protocol_version": "fake_version",
        }
        returnable.update(**ret)
        return returnable

    return connected


@pytest.fixture
def mock_interactive_session_block():
    with patch.multiple(
        "anyscale.connect_utils.start_interactive_session.StartInteractiveSessionBlock",
        __init__=Mock(return_value=None),
    ):
        interactive_session_block = StartInteractiveSessionBlock()
        interactive_session_block.api_client = Mock()
        interactive_session_block.log = Mock()
        interactive_session_block.block_label = ""
        return interactive_session_block


@pytest.mark.parametrize(
    "interactive_sessions_resp",
    [Mock(results=[]), Mock(results=[Mock(id="mock_job_id")])],
)
def test_log_interactive_session_info(
    mock_interactive_session_block, interactive_sessions_resp: Mock
):
    mock_interactive_session_block.api_client.list_decorated_interactive_sessions_api_v2_decorated_interactive_sessions_get = Mock(
        return_value=interactive_sessions_resp
    )
    mock_interactive_session_block._log_interactive_session_info(
        "mock_cluster_id",
        "mock_project_id",
        interactive_session_name="mock_interactive_session_name",
    )
    if not len(interactive_sessions_resp.results):
        mock_interactive_session_block.log.warning.assert_called_once_with(ANY)
    else:
        mock_interactive_session_block.log.warning.assert_not_called()


def test_acquire_session_lock_success(mock_interactive_session_block):
    """
    Test successful call to _acquire_session_lock.
    """
    mock_interactive_session_block._get_connect_params = Mock(
        return_value=("mock_session_url", "mock_secure", "mock_metadata")
    )
    info = {"num_clients": 0}
    mock_interactive_session_block._ray = Mock()
    mock_interactive_session_block._ray.util.connect = Mock(return_value=info)
    mock_interactive_session_block._ray.util.disconnect = Mock()
    mock_interactive_session_block._dynamic_check = Mock()
    mock_check_required_ray_version = Mock()
    mock_session = Mock()
    mock_session.name = "mock_session_name"
    mock_job_config = Mock()

    with patch.multiple(
        "anyscale.connect_utils.start_interactive_session",
        check_required_ray_version=mock_check_required_ray_version,
    ), patch.multiple(
        "inspect",
        getfullargspec=Mock(return_value=Mock(kwonlyargs=["ray_init_kwargs"])),
    ):
        assert (
            mock_interactive_session_block._acquire_session_lock(
                mock_session, 0, True, False, {}, job_config=mock_job_config
            )
            == info
        )
    mock_interactive_session_block._get_connect_params.assert_called_with(
        mock_session, True
    )
    mock_interactive_session_block._ray.util.connect.assert_called_with(
        "mock_session_url",
        connection_retries=0,
        ignore_version=True,
        job_config=mock_job_config,
        metadata="mock_metadata",
        secure="mock_secure",
        ray_init_kwargs={"logging_level": logging.ERROR},
    )
    mock_interactive_session_block._dynamic_check.assert_called_with(info, False)


@pytest.mark.parametrize("enable_multiple_clients", [True, False])
def test_acquire_session_multiple_clients(
    mock_interactive_session_block, enable_multiple_clients: bool
):
    mock_session = Mock()
    mock_interactive_session_block._ray = Mock()
    mock_interactive_session_block._ray.__version__ = "mock_ray_version"
    mock_interactive_session_block._ray.__commit__ = "mock_ray_commit"
    mock_interactive_session_block._ray.util.connect = Mock(
        side_effect=_connected(mock_interactive_session_block._ray, {"num_clients": 2})
    )

    def disconnected(*a: Any, **kw: Any) -> None:
        mock_interactive_session_block._ray.util.client.ray.is_connected.return_value = (
            False
        )

    # Emulate session lock failure.
    mock_interactive_session_block._ray.util.disconnect.side_effect = disconnected
    mock_interactive_session_block._get_connect_params = Mock(
        return_value=("mock_session_url", "mock_secure", "mock_metadata")
    )
    mock_interactive_session_block._dynamic_check = Mock()

    with patch.multiple(
        "anyscale.connect_utils.start_interactive_session",
        check_required_ray_version=Mock(),
    ):
        mock_interactive_session_block._acquire_session_lock(
            mock_session,
            0,
            True,
            False,
            {},
            job_config=Mock(),
            allow_multiple_clients=enable_multiple_clients,
        )

    assert mock_interactive_session_block._ray.util.connect.call_count == 2
    if enable_multiple_clients:
        # Should be connected successfully if ALLOW_MULTIPLE_CLIENTS = 1
        assert (
            mock_interactive_session_block._ray.util.client.ray.is_connected.return_value
        )
    else:
        # Should not be connected if ALLOW_MULTIPLE_CLIENTS = 0
        assert (
            not mock_interactive_session_block._ray.util.client.ray.is_connected.return_value
        )


@pytest.mark.parametrize(
    "ray_info_resp",
    [
        "",
        json.dumps(
            {"ray_commit": "mock_ray_commit", "ray_version": "mock_ray_version"}
        ).encode(),
    ],
)
def test_acquire_session_lock_failure(
    mock_interactive_session_block, ray_info_resp: str
):
    """
    Test _acquire_session_lock raises error when connection exception.
    """
    mock_interactive_session_block._get_connect_params = Mock(side_effect=RuntimeError)
    mock_interactive_session_block._subprocess = Mock()
    mock_interactive_session_block._subprocess.check_output = Mock(
        return_value=ray_info_resp
    )
    mock_check_required_ray_version = Mock()
    mock_interactive_session_block._ray = Mock()
    mock_interactive_session_block._ray.__version__ = Mock()
    mock_interactive_session_block._ray.__commit__ = Mock()

    mock_session = Mock()
    mock_session.name = "mock_session_name"

    with patch.multiple(
        "anyscale.connect_utils.start_interactive_session",
        check_required_ray_version=mock_check_required_ray_version,
    ):
        with pytest.raises(RuntimeError):
            mock_interactive_session_block._acquire_session_lock(
                mock_session, 0, True, False, {}
            )
    if ray_info_resp:
        mock_check_required_ray_version.assert_called_with(
            mock_interactive_session_block.log,
            mock_interactive_session_block._ray.__version__,
            mock_interactive_session_block._ray.__commit__,
            "mock_ray_version",
            "mock_ray_commit",
            False,
        )


@pytest.mark.parametrize(
    "connect_url",
    [None, "connect-ses-id.anyscale-prod-k8wcxpg-0000.anyscale-test-production.com"],
)
@pytest.mark.parametrize(
    "jupyter_notebook_url",
    [
        None,
        "https://dashboard-ses-id.anyscale-prod-k8wcxpg-0000.anyscale-test-production.com/jupyter/lab?token=mock_access_token",
    ],
)
@pytest.mark.parametrize("secure", [True, False])
def test_get_connect_params(
    mock_interactive_session_block,
    connect_url: Optional[str],
    jupyter_notebook_url: Optional[str],
    secure: bool,
):
    mock_session = Mock()
    mock_session.name = "mock_session_name"
    mock_session.connect_url = connect_url
    mock_session.jupyter_notebook_url = jupyter_notebook_url
    mock_session.access_token = "mock_access_token"

    if not connect_url and not jupyter_notebook_url:
        with pytest.raises(AssertionError):
            mock_interactive_session_block._get_connect_params(mock_session, secure)
        return

    (
        connect_url_output,
        secure_output,
        metadata_output,
    ) = mock_interactive_session_block._get_connect_params(mock_session, secure)

    assert secure_output == secure
    if connect_url:
        assert connect_url_output == connect_url
        assert metadata_output == [("cookie", "anyscale-token=mock_access_token")]
    elif secure:
        assert connect_url_output == (
            jupyter_notebook_url.split("/")[2].lower() if jupyter_notebook_url else None
        )
        assert metadata_output == [
            ("cookie", "anyscale-token=mock_access_token"),
            ("port", "10001"),
        ]
    else:
        assert connect_url_output == (
            jupyter_notebook_url.split("/")[2].lower() + ":8081"
            if jupyter_notebook_url
            else None
        )
        assert metadata_output == [
            ("cookie", "anyscale-token=mock_access_token"),
            ("port", "10001"),
        ]


@pytest.mark.parametrize("python_version", ["3", "2"])
def test_dynamic_check(mock_interactive_session_block, python_version: str):
    mock_interactive_session_block._ray = Mock()
    mock_interactive_session_block._ray.__version__ = "1.8.0"
    mock_interactive_session_block._ray.__commit__ = "mock_commit_id"
    mock_check_required_ray_version = Mock()
    mock_detect_python_minor_version = Mock(return_value="3")
    mock_info = {
        "ray_version": "1.8.0",
        "ray_commit": "mock_commit_id",
        "python_version": python_version,
    }

    with patch.multiple(
        "anyscale.connect_utils.start_interactive_session",
        check_required_ray_version=mock_check_required_ray_version,
        detect_python_minor_version=mock_detect_python_minor_version,
    ):
        if python_version != "3":
            with pytest.raises(AssertionError):
                mock_interactive_session_block._dynamic_check(mock_info, False)
        else:
            mock_interactive_session_block._dynamic_check(mock_info, False)
    mock_check_required_ray_version.assert_called_with(
        mock_interactive_session_block.log,
        mock_interactive_session_block._ray.__version__,
        mock_interactive_session_block._ray.__commit__,
        mock_info["ray_version"],
        mock_info["ray_commit"],
        False,
    )
    mock_detect_python_minor_version.assert_called_with()


@pytest.mark.parametrize("is_connected_resp", [True, False])
@pytest.mark.parametrize(
    "host_name",
    [
        None,
        "https://dashboard-ses-id.anyscale-prod-k8wcxpg-0000.anyscale-test-production.com",
    ],
)
@pytest.mark.parametrize(
    "jupyter_notebook_url",
    [
        None,
        "https://dashboard-ses-id.anyscale-prod-k8wcxpg-0000.anyscale-test-production.com/jupyter/lab?token=mock_access_token",
    ],
)
def test_check_connection(
    mock_interactive_session_block,
    is_connected_resp: bool,
    host_name: Optional[str],
    jupyter_notebook_url: Optional[str],
):
    mock_interactive_session_block._ray = Mock()
    mock_interactive_session_block._ray.util.client.ray.is_connected = Mock(
        return_value=is_connected_resp
    )
    mock_session = Mock(
        id="mock_session_id",
        host_name=host_name,
        jupyter_notebook_url=jupyter_notebook_url,
    )

    if not is_connected_resp:
        with pytest.raises(RuntimeError):
            mock_interactive_session_block._check_connection(mock_session)
    else:
        mock_interactive_session_block._check_connection(mock_session)

    mock_interactive_session_block._ray.util.client.ray.is_connected.assert_called_with()


@pytest.mark.parametrize("initial_scale", [[], [{"mock_key": 1}]])
def test_init(initial_scale: INITIAL_SCALE_TYPE):
    mock_get_interactive_shell_frame = Mock(return_value=None)
    mock_cluster = Mock()
    mock_connection_info = {
        "python_version": "mock_python_version",
        "ray_version": "mock_ray_version",
        "ray_commit": "mock_ray_commit",
        "protocol_version": "mock_protocol_version",
        "num_clients": "mock_num_clients",
    }
    mock_acquire_session_lock = Mock(return_value=mock_connection_info)
    mock_check_connection = Mock()
    mock_ray = Mock()
    mock_ray.__version__ = mock_connection_info["ray_version"]
    mock_ray.__commit__ = mock_connection_info["ray_commit"]
    mock_ray.job_config.JobConfig = Mock
    mock_job_config = Mock()
    mock_log_interactive_session_info = Mock()
    with patch.multiple(
        "anyscale.connect_utils.start_interactive_session",
        get_auth_api_client=Mock(return_value=Mock(),),
        _get_interactive_shell_frame=mock_get_interactive_shell_frame,
    ), patch.multiple(
        "anyscale.connect_utils.start_interactive_session.StartInteractiveSessionBlock",
        _acquire_session_lock=mock_acquire_session_lock,
        _check_connection=mock_check_connection,
        _log_interactive_session_info=mock_log_interactive_session_info,
    ):
        interactive_session_block_output = StartInteractiveSessionBlock(
            mock_cluster,
            mock_job_config,
            False,
            initial_scale,
            False,
            None,
            {},
            True,
            False,
            mock_ray,
            Mock(),
        )

    assert (
        interactive_session_block_output.anyscale_client_context.dashboard_url
        == mock_cluster.ray_dashboard_url
    )
    assert (
        interactive_session_block_output.anyscale_client_context.python_version
        == mock_connection_info["python_version"]
    )
    assert (
        interactive_session_block_output.anyscale_client_context.ray_version
        == mock_connection_info["ray_version"]
    )
    assert (
        interactive_session_block_output.anyscale_client_context.ray_commit
        == mock_connection_info["ray_commit"]
    )
    assert (
        interactive_session_block_output.anyscale_client_context.protocol_version
        == mock_connection_info["protocol_version"]
    )
    assert (
        interactive_session_block_output.anyscale_client_context._num_clients
        == mock_connection_info["num_clients"]
    )
    assert interactive_session_block_output.connection_info == mock_connection_info

    mock_acquire_session_lock.assert_called_with(
        mock_cluster,
        connection_retries=10,
        secure=True,
        ignore_version_check=False,
        ray_init_kwargs={},
        job_config=mock_job_config,
        allow_multiple_clients=False,
    )
    mock_check_connection.assert_called_with(mock_cluster)
    mock_get_interactive_shell_frame.assert_called_with()
    mock_log_interactive_session_info.assert_called_with(
        mock_cluster.id,
        mock_cluster.project_id,
        mock_job_config.metadata.get("job_name"),
    )

    if len(initial_scale):
        mock_ray.autoscaler.sdk.request_resources.assert_called_with(
            bundles=initial_scale
        )

    expected_context = AnyscaleClientContext(
        anyscale_cluster_info=AnyscaleClientConnectResponse(cluster_id="cluster-1"),
        python_version=mock_connection_info["python_version"],
        _num_clients=mock_connection_info["num_clients"],
        ray_version=mock_ray.__version__,
        ray_commit=mock_ray.__commit__,
        protocol_version=mock_connection_info["protocol_version"],
        dashboard_url=mock_cluster.ray_dashboard_url,
    )
    assert interactive_session_block_output.anyscale_client_context == expected_context



import os
from typing import Any

import boto3

from anyscale.authenticate import get_auth_api_client


"""anyscale/experimental_integrations.py: Experimental util functions for W&B integration and secret store prototypes."""

WANDB_API_KEY_NAME = "WANDB_API_KEY_NAME"


def get_aws_secret(secret_name: str, **kwargs) -> str:
    """
    Get secret value from AWS secrets manager.

    Arguments:
        secret_name: Key of your secret
        kwargs: Optional credentials passed in to authenticate instance
    """
    client = boto3.client("secretsmanager", **kwargs)
    response = client.get_secret_value(SecretId=secret_name)

    # Depending on whether the secret is a string or binary, one of these fields will be populated.
    if "SecretString" in response:
        secret = response.pop("SecretString")
    else:
        secret = response.pop("SecretBinary")

    # TODO(nikita): Return secret as a ProtectedString
    return secret


def wandb_setup_api_key_hook() -> str:
    """
    Returns W&B API key based on key set in WANDB_API_KEY_NAME in
    AWS secrets manager. This returns the API key in plain text,
    so take care to not save the output in any logs.

    Assumes instance is running with correct IAM role, so credentials
    don't have to be passed to access secrets manager.
    """
    secret_name = os.environ.get(WANDB_API_KEY_NAME)
    production_job_id = os.environ.get("ANYSCALE_PRODUCTION_JOB_ID")
    if not production_job_id:
        raise Exception(
            "ANYSCALE_PRODUCTION_JOB_ID environment variable not set. The W&B integration is "
            "currently only supported for production jobs."
        )

    api_client = get_auth_api_client(log_output=False).api_client
    production_job = api_client.get_job_api_v2_decorated_ha_jobs_production_job_id_get(
        production_job_id
    ).result

    if not secret_name:
        user_id = production_job.creator_id
        secret_name = f"wandb_api_key_{user_id}"

    # Get cloud from production job to verify job is using an AWS cloud and
    # to pass the region to the boto3 api.
    cluster_id = (
        production_job.state.cluster.id
        if (production_job.state and production_job.state.cluster)
        else None
    )
    if cluster_id:
        cluster = api_client.get_session_api_v2_sessions_session_id_get(
            cluster_id
        ).result
        cloud_id = cluster.cloud_id
        cloud = api_client.get_cloud_api_v2_clouds_cloud_id_get(cloud_id).result
    else:
        raise Exception(
            f"Unable to find cluster for production job {production_job_id}"
        )

    if cloud.provider != "AWS":
        raise Exception(
            "The Anyscale W&B integration is currently only supported for AWS clouds."
        )
    region = cloud.region

    secret = get_aws_secret(secret_name, region_name=region)
    return secret


def wandb_send_run_info_hook(run: Any) -> None:
    try:
        import wandb
    except ImportError:
        raise Exception("Unable to import wandb.")

    assert isinstance(
        run, wandb.sdk.wandb_run.Run
    ), "`run` argument must be of type wandb.sdk.wandb_run.Run"

    api_client = get_auth_api_client(log_output=False).api_client

    if os.environ.get("ANYSCALE_PRODUCTION_JOB_ID"):
        production_job_id = os.environ.get("ANYSCALE_PRODUCTION_JOB_ID")
        api_client.update_wandb_run_values_api_v2_experimental_integrations_update_wandb_run_values_production_job_id_get(
            production_job_id=production_job_id,
            wandb_entity=run.entity,
            wandb_project=run.project,
            wandb_group=run.group,
        )



import sys
from time import sleep, time
from typing import Any, Dict, Optional

from anyscale.api import configure_tcp_keepalive
from anyscale.sdk.anyscale_client.models.cluster_compute_config import ClusterComputeConfig
from anyscale.sdk.anyscale_client.models.create_cluster_compute import CreateClusterCompute
from anyscale.utils.runtime_env import upload_and_rewrite_working_dir
import anyscale.sdk.anyscale_client as anyscale_client
from anyscale.sdk.anyscale_client.api.default_api import DefaultApi
from anyscale.version import __version__ as version
from anyscale.sdk.anyscale_client.models.create_cluster_environment import (
    CreateClusterEnvironment,
)
from anyscale.sdk.anyscale_client.models.cluster import Cluster
from anyscale.sdk.anyscale_client.models.cluster_environment_build import (
    ClusterEnvironmentBuild,
)
from anyscale.shared_anyscale_utils.headers import RequestHeaders
from anyscale.cli_logger import BlockLogger
from anyscale.util import get_endpoint
from anyscale.authenticate import AuthenticationBlock
from anyscale_client.models import ClusterEnvironmentBuildStatus

def _upload_and_rewrite_working_dir_in_create_production_job(create_production_job):
    if create_production_job.config is not None:
        runtime_env = create_production_job.config.get("runtime_env")
        if runtime_env is not None:
            new_runtime_env = upload_and_rewrite_working_dir(runtime_env)
            create_production_job.config["runtime_env"] = new_runtime_env

# Build states that are in progress
IN_PROGRESS_BUILD_STATES = [ClusterEnvironmentBuildStatus.IN_PROGRESS, ClusterEnvironmentBuildStatus.PENDING]

class AnyscaleSDK(DefaultApi):  # type: ignore
    def __init__(
        self, auth_token: Optional[str] = None, host: str = "https://api.anyscale.com",
    ):
        # Special handling for `console.` host names
        if host.startswith("https://console."):
            host = host.replace("console.", "api.", 1)

        # Adds base path "v0" for API endpoints
        endpoint_url = host.rstrip("/") + "/v0"
        configuration = anyscale_client.Configuration(host=endpoint_url)
        configuration.connection_pool_maxsize = 100
        if auth_token is None:
            auth_token, _ = AuthenticationBlock._load_credentials()
        api_client = anyscale_client.ApiClient(
            configuration, cookie=f"cli_token={auth_token}"
        )
        configure_tcp_keepalive(api_client)
        api_client.set_default_header(RequestHeaders.CLIENT, "SDK")
        api_client.set_default_header(RequestHeaders.CLIENT_VERSION, version)

        self.log: BlockLogger = BlockLogger()

        super(AnyscaleSDK, self).__init__(api_client)

    def build_cluster_environment(
        self,
        create_cluster_environment: CreateClusterEnvironment,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        log_output: bool = False,
    ) -> ClusterEnvironmentBuild:
        """
        Creates a new Cluster Environment and waits for build to complete.

        If a Cluster Environment with the same name already exists, this will
        create an updated build of that environment.

        Args:
            create_cluster_environment - CreateClusterEnvironment object
            poll_rate_seconds - seconds to wait when polling build operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for build operation to complete before timing out; defaults to no timeout

        Returns:
            Newly created ClusterEnvironmentBuild object

        Raises:
            Exception if building Cluster Environment failed or timed out
        """

        self.log.log_output = log_output
        cluster_environments = self.search_cluster_environments(
            {
                "name": {"equals": create_cluster_environment.name},
                "paging": {"count": 1},
            }
        ).results

        if not cluster_environments:
            self.log.info(
                f"Creating new cluster environment {create_cluster_environment.name}"
            )
            cluster_environment = self.create_cluster_environment(
                create_cluster_environment
            ).result
            build = self.list_cluster_environment_builds(
                cluster_environment.id
            ).results[0]
            build_operation_id = build.id
        else:
            self.log.info(
                f"Building new revision of cluster environment {create_cluster_environment.name}"
            )
            cluster_environment = cluster_environments[0]
            build = self.create_cluster_environment_build(
                {
                    "cluster_environment_id": cluster_environment.id,
                    "config_json": create_cluster_environment.config_json,
                }
            ).result
            build_operation_id = build.id

        return self.wait_for_cluster_environment_build_operation(
            build_operation_id, poll_rate_seconds, timeout_seconds, log_output
        )

    def wait_for_cluster_environment_build_operation(
        self,
        operation_id: str,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        log_output: bool = False,
    ) -> ClusterEnvironmentBuild:
        """
        Waits for a Cluster Environment Build operation to complete.

        Args:
            operation_id - ID of the Cluster Environment Build operation
            poll_rate_seconds - seconds to wait when polling build operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for build operation to complete before timing out; defaults to no timeout

        Returns:
            ClusterEnvironmentBuild object

        Raises:
            Exception if building Cluster Environment fails or times out.
        """

        self.log.log_output = log_output
        timeout = time() + timeout_seconds if timeout_seconds else None

        operation = self.get_build(operation_id)
        url = get_endpoint(
            f"configurations/app-config-details/{operation_id}"
        )
        self.log.info(
            f"Waiting for cluster environment to build. View progress at {url}."
        )
        self.log.info(f"status: {operation.result.status}")
        while operation.result.status in IN_PROGRESS_BUILD_STATES:
            if timeout and time() > timeout:
                raise Exception(
                    f"Building Cluster Environment timed out after {timeout_seconds} seconds."
                )

            sleep(poll_rate_seconds)
            self.log.info(f"status: {operation.result.status}")
            operation = self.get_build(
                operation_id
            )
        if operation.result.status != ClusterEnvironmentBuildStatus.SUCCEEDED:
            raise Exception(
                f"Failed to build Cluster Environment, you can check the full logs at: {url}."
            )
        else:
            self.log.info("Cluster environment successfully finished building.")
            return self.get_build(
                operation_id
            ).result

    def launch_cluster(
        self,
        project_id: Optional[str],
        cluster_name: str,
        cluster_environment_build_id: Optional[str] = None,
        cluster_compute_id: Optional[str] = None,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        idle_timeout_minutes: Optional[int] = None,
        cluster_compute_config: Optional[ClusterComputeConfig] = None,
    ) -> Cluster:
        """
        Starts a Cluster in the specified Project.
        If a Cluster with the specified name already exists, we will update that Cluster.
        Otherwise, a new Cluster will be created.

        Args:
            project_id - ID of the Project the Cluster belongs to or None to launch cluster without a project.
            cluster_name - Name of the Cluster
            cluster_environment_build_id - Cluster Environment Build to start this Cluster with
                                           If none, uses a default cluster_environment_build
            cluster_compute_id - Cluster Compute to start this Cluster with
                                 If none, it checks if `cluster_compute_config` is specified. 
                                 If `cluster_compute_config` is none, it uses a default cluster_compute.
            poll_rate_seconds - seconds to wait when polling Cluster operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for Cluster operation to complete before timing out; defaults to no timeout
            idle_timeout_minutes - Idle timeout (in minutes), after which the Cluster is terminated
            cluster_compute_config - One-off Cluster Compute that this Cluster will use.

        Returns:
            Cluster object

        Raises:
            Exception if starting Cluster fails or times out
        """

        if not project_id:
            default_project = self.get_default_project().result
            search_project_id = default_project.id
        else:
            search_project_id = project_id

        clusters = self.search_clusters(
            {"project_id": search_project_id, "name": {"equals": cluster_name}}
        ).results

        if clusters:
            cluster = clusters[0]
        else:
            if not cluster_environment_build_id:
                # Use default cluster environment build when starting cluster without specifying an id
                py_version = "".join(str(x) for x in sys.version_info[0:2])
                if py_version not in ["36", "37", "38"]:
                    raise ValueError(
                        "No default cluster env for python version {}. Please use a version of python between 3.6 and 3.8.".format(
                            py_version
                        )
                    )
                import ray

                ray_version = ray.__version__
                if "dev0" in ray_version:
                    raise ValueError(
                        f"Your locally installed Ray version is {ray_version}. "
                        "There is no default cluster environments for nightly versions of Ray."
                    )
                cluster_environment_build_id = self.get_default_cluster_environment_build(
                    f"py{py_version}", ray_version
                ).result.id

            if not cluster_compute_id:
                if cluster_compute_config:
                    cluster_compute_id = self.create_cluster_compute(CreateClusterCompute(config=cluster_compute_config, anonymous=True)).result.id
                else:
                    # Use default cluster compute when starting cluster without specifying an id
                    cluster_compute_id = self.get_default_cluster_compute().result.id

            create_cluster_payload: Dict[str, Any] = {
                "name": cluster_name,
                "project_id": project_id,
                "cluster_environment_build_id": cluster_environment_build_id,
                "cluster_compute_id": cluster_compute_id,
            }
            if idle_timeout_minutes:
                create_cluster_payload["idle_timeout_minutes"] = idle_timeout_minutes

            cluster = self.create_cluster(create_cluster_payload).result

        start_operation = self.start_cluster(
            cluster.id,
            {
                "cluster_environment_build_id": cluster_environment_build_id,
                "cluster_compute_id": cluster_compute_id,
            },
        ).result

        return self.wait_for_cluster_operation(
            start_operation.id, poll_rate_seconds, timeout_seconds
        )

    def launch_cluster_with_new_cluster_environment(
        self,
        project_id: Optional[str],
        cluster_name: str,
        create_cluster_environment: CreateClusterEnvironment,
        cluster_compute_id: Optional[str] = None,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
        idle_timeout_minutes: Optional[int] = None,
        cluster_compute_config: Optional[ClusterComputeConfig] = None,
    ) -> Cluster:
        """
        Builds a new Cluster Environment, then starts a Cluster in the specified Project with the new build.
        If a Cluster with the specified name already exists, we will update that Cluster.
        Otherwise, a new Cluster will be created.

        Args:
            project_id - ID of the Project the Cluster belongs to or None to launch a cluster without a project.
            cluster_name - Name of the Cluster
            create_cluster_environment - CreateClusterEnvironment object
            cluster_compute_id - Cluster Compute to start this Cluster with
                                 If none, it checks if `cluster_compute_config` is specified. 
                                 If `cluster_compute_config` is  none, it uses a default cluster_compute.
            poll_rate_seconds - seconds to wait when polling for operations; defaults to 15
            timeout_seconds - maximum number of seconds to wait for each operation to complete before timing out; defaults to no timeout
            idle_timeout_minutes - Idle timeout (in minutes), after which the Cluster is terminated
            cluster_compute_config - One-off Cluster Compute that this Cluster will use.

        Returns:
            Cluster object

        Raises:
            Exception if building the new Cluster Environment fails or starting the Cluster fails.
        """

        cluster_environment_build = self.build_cluster_environment(
            create_cluster_environment, poll_rate_seconds, timeout_seconds
        )

        return self.launch_cluster(
            project_id,
            cluster_name,
            cluster_environment_build.id,
            cluster_compute_id,
            poll_rate_seconds,
            timeout_seconds,
            idle_timeout_minutes,
            cluster_compute_config,
        )

    def wait_for_cluster_operation(
        self,
        operation_id: str,
        poll_rate_seconds: int = 15,
        timeout_seconds: Optional[int] = None,
    ) -> Cluster:
        """
        Waits for a Cluster operation to complete, most commonly used when starting, terminating, or updating a Cluster.

        Args:
            operation_id - ID of the Cluster Operation
            poll_rate_seconds - seconds to wait when polling build operation status; defaults to 15
            timeout_seconds - maximum number of seconds to wait for build operation to complete before timing out; defaults to no timeout

        Returns:
            Cluster object when the operation completes successfully

        Raises:
            Exception if building Cluster operation fails or times out
        """

        timeout = time() + timeout_seconds if timeout_seconds else None

        operation = self.get_cluster_operation(operation_id).result
        while not operation.completed:
            if timeout and time() > timeout:
                raise Exception(
                    f"Cluster start up timed out after {timeout_seconds} seconds."
                )

            sleep(poll_rate_seconds)
            operation = self.get_cluster_operation(operation_id).result

        if operation.result.error:
            raise Exception("Failed to start Cluster", operation.result.error)
        else:
            return self.get_cluster(operation.cluster_id).result

    def fetch_actor_logs(self, actor_id: str) -> str:
        """
        Retrieves logs for an Actor.
        This function may take several minutes if the Cluster this Actor ran on has been terminated.

        Args:
            actor_id - ID of the Actor

        Returns
            Log output for the Actor as a string

        Raises
            Exception if we fail to fetch logs
        """

        # Timeout in 5 minutes
        timeout = time() + 300

        logs = self.get_actor_logs(actor_id).result
        while not logs.ready:
            if time() > timeout:
                raise Exception(f"Failed to get logs from Actor {actor_id}")

            sleep(15)
            logs = self.get_actor_logs(actor_id).result

        return str(logs.logs)

    def fetch_job_logs(self, job_id: str) -> str:
        """
        Retrieves logs for a Ray Job or Interactive Session.
        This function may take several minutes if the Cluster this Job ran on has been terminated.

        Args:
            job_id - ID of the Job

        Returns
            Log output for the Job as a string

        Raises
            Exception if we fail to fetch logs
        """

        # Timeout in 5 minutes
        timeout = time() + 300

        logs = self.get_job_logs(job_id).result

        while not logs.ready:
            if time() > timeout:
                raise Exception(f"Failed to get logs from Job {job_id}")

            sleep(15)
            logs = self.get_job_logs(job_id).result

        return str(logs.logs)

    def get_production_job_logs(self, job_id: str) -> str:
        """
        Retrieves logs for latest run of a Production Job.

        Args:
            job_id - ID of the Job

        Returns
            Log output for the Job as a string

        Raises
            Exception if we fail to fetch logs
        """

        latest_job_run_id = self.get_production_job(
            production_job_id=job_id
        ).result.last_job_run_id
        if not latest_job_run_id:
            raise Exception(f"Production Job {job_id} does not have a job run.")

        return self.get_job_logs(latest_job_run_id)

    def fetch_production_job_logs(self, job_id: str) -> str:
        """
        Retrieves logs for latest run of a Production Job.
        Waits 5 min for logs to be ready.
        This function may take several minutes if the Cluster this Job ran on has been terminated.

        Args:
            job_id - ID of the Job

        Returns
            Log output for the Job as a string

        Raises
            Exception if we fail to fetch logs
        """

        # Timeout in 5 minutes
        timeout = time() + 300

        logs = self.get_production_job_logs(job_id).result

        while not logs.ready:
            if time() > timeout:
                raise Exception(f"Failed to get logs from Job {job_id}")

            sleep(15)
            logs = self.get_production_job_logs(job_id).result

        return str(logs.logs)

    def create_job(self, create_production_job, **kwargs):
        _upload_and_rewrite_working_dir_in_create_production_job(create_production_job)
        return super().create_job(create_production_job, **kwargs)

    def create_service(self, create_production_service, **kwargs):
        _upload_and_rewrite_working_dir_in_create_production_job(
            create_production_service
        )
        return super().create_service(create_production_service, **kwargs)

    def apply_service(self, create_production_service, **kwargs):
        _upload_and_rewrite_working_dir_in_create_production_job(
            create_production_service
        )
        return super().apply_service(create_production_service, **kwargs)

    def get_cluster_token(self, project_id: str, cluster_name: str) -> str:
        """
        Retrieves cluster token for a Cluster.
        This cluster token can be used to authenticate Anyscale services (eg: serve, jupyter, grafana)

        Args:
            project_id - ID of the Project the Cluster belongs to
            cluster_name - Name of the Cluster

        Returns
            string containing cluster token

        Raises
            Exception if invalid cluster name
        """
        clusters = self.search_clusters(
            {"project_id": project_id, "name": {"equals": cluster_name}}
        ).results

        if clusters:
            cluster = clusters[0]
        else:
            raise Exception(
                (
                    f"Cluster named {cluster_name} does not exist in Project {project_id} "
                    "or this user does not have read permissions for it."
                )
            )
        cluster_token = cluster.access_token
        return cluster_token



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class ProductionJobConfig(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'entrypoint': 'str',
        'runtime_env': 'RayRuntimeEnvConfig',
        'build_id': 'str',
        'compute_config_id': 'str',
        'compute_config': 'ClusterComputeConfig',
        'max_retries': 'int',
        'runtime_env_config': 'RayRuntimeEnvConfig'
    }

    attribute_map = {
        'entrypoint': 'entrypoint',
        'runtime_env': 'runtime_env',
        'build_id': 'build_id',
        'compute_config_id': 'compute_config_id',
        'compute_config': 'compute_config',
        'max_retries': 'max_retries',
        'runtime_env_config': 'runtime_env_config'
    }

    def __init__(self, entrypoint=None, runtime_env=None, build_id=None, compute_config_id=None, compute_config=None, max_retries=5, runtime_env_config=None, local_vars_configuration=None):  # noqa: E501
        """ProductionJobConfig - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._entrypoint = None
        self._runtime_env = None
        self._build_id = None
        self._compute_config_id = None
        self._compute_config = None
        self._max_retries = None
        self._runtime_env_config = None
        self.discriminator = None

        self.entrypoint = entrypoint
        if runtime_env is not None:
            self.runtime_env = runtime_env
        self.build_id = build_id
        self.compute_config_id = compute_config_id
        if compute_config is not None:
            self.compute_config = compute_config
        if max_retries is not None:
            self.max_retries = max_retries
        if runtime_env_config is not None:
            self.runtime_env_config = runtime_env_config

    @property
    def entrypoint(self):
        """Gets the entrypoint of this ProductionJobConfig.  # noqa: E501

        A script that will be run to start your job. This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'  # noqa: E501

        :return: The entrypoint of this ProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._entrypoint

    @entrypoint.setter
    def entrypoint(self, entrypoint):
        """Sets the entrypoint of this ProductionJobConfig.

        A script that will be run to start your job. This command will be run in the root directory of the specified runtime env. Eg. 'python script.py'  # noqa: E501

        :param entrypoint: The entrypoint of this ProductionJobConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and entrypoint is None:  # noqa: E501
            raise ValueError("Invalid value for `entrypoint`, must not be `None`")  # noqa: E501

        self._entrypoint = entrypoint

    @property
    def runtime_env(self):
        """Gets the runtime_env of this ProductionJobConfig.  # noqa: E501

        A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.  # noqa: E501

        :return: The runtime_env of this ProductionJobConfig.  # noqa: E501
        :rtype: RayRuntimeEnvConfig
        """
        return self._runtime_env

    @runtime_env.setter
    def runtime_env(self, runtime_env):
        """Sets the runtime_env of this ProductionJobConfig.

        A ray runtime env json. Your entrypoint will be run in the environment specified by this runtime env.  # noqa: E501

        :param runtime_env: The runtime_env of this ProductionJobConfig.  # noqa: E501
        :type: RayRuntimeEnvConfig
        """

        self._runtime_env = runtime_env

    @property
    def build_id(self):
        """Gets the build_id of this ProductionJobConfig.  # noqa: E501

        The id of the cluster env build. This id will determine the docker image your job is run on.  # noqa: E501

        :return: The build_id of this ProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._build_id

    @build_id.setter
    def build_id(self, build_id):
        """Sets the build_id of this ProductionJobConfig.

        The id of the cluster env build. This id will determine the docker image your job is run on.  # noqa: E501

        :param build_id: The build_id of this ProductionJobConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `build_id`, must not be `None`")  # noqa: E501

        self._build_id = build_id

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this ProductionJobConfig.  # noqa: E501

        The id of the compute configuration that you want to use. This id will specify the resources required for your job  # noqa: E501

        :return: The compute_config_id of this ProductionJobConfig.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this ProductionJobConfig.

        The id of the compute configuration that you want to use. This id will specify the resources required for your job  # noqa: E501

        :param compute_config_id: The compute_config_id of this ProductionJobConfig.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def compute_config(self):
        """Gets the compute_config of this ProductionJobConfig.  # noqa: E501

        One-off compute that the cluster will use.  # noqa: E501

        :return: The compute_config of this ProductionJobConfig.  # noqa: E501
        :rtype: ClusterComputeConfig
        """
        return self._compute_config

    @compute_config.setter
    def compute_config(self, compute_config):
        """Sets the compute_config of this ProductionJobConfig.

        One-off compute that the cluster will use.  # noqa: E501

        :param compute_config: The compute_config of this ProductionJobConfig.  # noqa: E501
        :type: ClusterComputeConfig
        """

        self._compute_config = compute_config

    @property
    def max_retries(self):
        """Gets the max_retries of this ProductionJobConfig.  # noqa: E501

        The number of retries this job will attempt on failure. Set to None to set infinite retries  # noqa: E501

        :return: The max_retries of this ProductionJobConfig.  # noqa: E501
        :rtype: int
        """
        return self._max_retries

    @max_retries.setter
    def max_retries(self, max_retries):
        """Sets the max_retries of this ProductionJobConfig.

        The number of retries this job will attempt on failure. Set to None to set infinite retries  # noqa: E501

        :param max_retries: The max_retries of this ProductionJobConfig.  # noqa: E501
        :type: int
        """

        self._max_retries = max_retries

    @property
    def runtime_env_config(self):
        """Gets the runtime_env_config of this ProductionJobConfig.  # noqa: E501

        DEPRECATED: Use runtime_env  # noqa: E501

        :return: The runtime_env_config of this ProductionJobConfig.  # noqa: E501
        :rtype: RayRuntimeEnvConfig
        """
        return self._runtime_env_config

    @runtime_env_config.setter
    def runtime_env_config(self, runtime_env_config):
        """Sets the runtime_env_config of this ProductionJobConfig.

        DEPRECATED: Use runtime_env  # noqa: E501

        :param runtime_env_config: The runtime_env_config of this ProductionJobConfig.  # noqa: E501
        :type: RayRuntimeEnvConfig
        """

        self._runtime_env_config = runtime_env_config

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ProductionJobConfig):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ProductionJobConfig):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class CreateExperimentalWorkspace(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'description': 'str',
        'cloud_id': 'str',
        'compute_config_id': 'str',
        'base_snapshot': 'str',
        'cluster_environment_build_id': 'str',
        'idle_timeout_minutes': 'int',
        'cloned_job_id': 'str',
        'cloned_workspace_id': 'str'
    }

    attribute_map = {
        'name': 'name',
        'description': 'description',
        'cloud_id': 'cloud_id',
        'compute_config_id': 'compute_config_id',
        'base_snapshot': 'base_snapshot',
        'cluster_environment_build_id': 'cluster_environment_build_id',
        'idle_timeout_minutes': 'idle_timeout_minutes',
        'cloned_job_id': 'cloned_job_id',
        'cloned_workspace_id': 'cloned_workspace_id'
    }

    def __init__(self, name=None, description=None, cloud_id=None, compute_config_id=None, base_snapshot=None, cluster_environment_build_id=None, idle_timeout_minutes=120, cloned_job_id=None, cloned_workspace_id=None, local_vars_configuration=None):  # noqa: E501
        """CreateExperimentalWorkspace - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._description = None
        self._cloud_id = None
        self._compute_config_id = None
        self._base_snapshot = None
        self._cluster_environment_build_id = None
        self._idle_timeout_minutes = None
        self._cloned_job_id = None
        self._cloned_workspace_id = None
        self.discriminator = None

        self.name = name
        if description is not None:
            self.description = description
        self.cloud_id = cloud_id
        self.compute_config_id = compute_config_id
        if base_snapshot is not None:
            self.base_snapshot = base_snapshot
        self.cluster_environment_build_id = cluster_environment_build_id
        if idle_timeout_minutes is not None:
            self.idle_timeout_minutes = idle_timeout_minutes
        if cloned_job_id is not None:
            self.cloned_job_id = cloned_job_id
        if cloned_workspace_id is not None:
            self.cloned_workspace_id = cloned_workspace_id

    @property
    def name(self):
        """Gets the name of this CreateExperimentalWorkspace.  # noqa: E501

        Name of the workspace to be created.  # noqa: E501

        :return: The name of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this CreateExperimentalWorkspace.

        Name of the workspace to be created.  # noqa: E501

        :param name: The name of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def description(self):
        """Gets the description of this CreateExperimentalWorkspace.  # noqa: E501

        Description of Workspace.  # noqa: E501

        :return: The description of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._description

    @description.setter
    def description(self, description):
        """Sets the description of this CreateExperimentalWorkspace.

        Description of Workspace.  # noqa: E501

        :param description: The description of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """

        self._description = description

    @property
    def cloud_id(self):
        """Gets the cloud_id of this CreateExperimentalWorkspace.  # noqa: E501

        The cloud id for the workspace  # noqa: E501

        :return: The cloud_id of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._cloud_id

    @cloud_id.setter
    def cloud_id(self, cloud_id):
        """Sets the cloud_id of this CreateExperimentalWorkspace.

        The cloud id for the workspace  # noqa: E501

        :param cloud_id: The cloud_id of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cloud_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cloud_id`, must not be `None`")  # noqa: E501

        self._cloud_id = cloud_id

    @property
    def compute_config_id(self):
        """Gets the compute_config_id of this CreateExperimentalWorkspace.  # noqa: E501

        The compute config id for the workspace  # noqa: E501

        :return: The compute_config_id of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._compute_config_id

    @compute_config_id.setter
    def compute_config_id(self, compute_config_id):
        """Sets the compute_config_id of this CreateExperimentalWorkspace.

        The compute config id for the workspace  # noqa: E501

        :param compute_config_id: The compute_config_id of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and compute_config_id is None:  # noqa: E501
            raise ValueError("Invalid value for `compute_config_id`, must not be `None`")  # noqa: E501

        self._compute_config_id = compute_config_id

    @property
    def base_snapshot(self):
        """Gets the base_snapshot of this CreateExperimentalWorkspace.  # noqa: E501

        Metadata on base snapshot  # noqa: E501

        :return: The base_snapshot of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._base_snapshot

    @base_snapshot.setter
    def base_snapshot(self, base_snapshot):
        """Sets the base_snapshot of this CreateExperimentalWorkspace.

        Metadata on base snapshot  # noqa: E501

        :param base_snapshot: The base_snapshot of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """

        self._base_snapshot = base_snapshot

    @property
    def cluster_environment_build_id(self):
        """Gets the cluster_environment_build_id of this CreateExperimentalWorkspace.  # noqa: E501

        The cluster environment build id for the cluster used by the workspace  # noqa: E501

        :return: The cluster_environment_build_id of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._cluster_environment_build_id

    @cluster_environment_build_id.setter
    def cluster_environment_build_id(self, cluster_environment_build_id):
        """Sets the cluster_environment_build_id of this CreateExperimentalWorkspace.

        The cluster environment build id for the cluster used by the workspace  # noqa: E501

        :param cluster_environment_build_id: The cluster_environment_build_id of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_environment_build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_environment_build_id`, must not be `None`")  # noqa: E501

        self._cluster_environment_build_id = cluster_environment_build_id

    @property
    def idle_timeout_minutes(self):
        """Gets the idle_timeout_minutes of this CreateExperimentalWorkspace.  # noqa: E501

        Idle timeout (in minutes), after which the Cluster is terminated. Idle time is defined as the time during which a Cluster is not running a user command (through 'anyscale exec' or the Web UI), and does not have an attached driver. Time spent running Jupyter commands, or commands run through ssh, is still considered 'idle'.  # noqa: E501

        :return: The idle_timeout_minutes of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: int
        """
        return self._idle_timeout_minutes

    @idle_timeout_minutes.setter
    def idle_timeout_minutes(self, idle_timeout_minutes):
        """Sets the idle_timeout_minutes of this CreateExperimentalWorkspace.

        Idle timeout (in minutes), after which the Cluster is terminated. Idle time is defined as the time during which a Cluster is not running a user command (through 'anyscale exec' or the Web UI), and does not have an attached driver. Time spent running Jupyter commands, or commands run through ssh, is still considered 'idle'.  # noqa: E501

        :param idle_timeout_minutes: The idle_timeout_minutes of this CreateExperimentalWorkspace.  # noqa: E501
        :type: int
        """

        self._idle_timeout_minutes = idle_timeout_minutes

    @property
    def cloned_job_id(self):
        """Gets the cloned_job_id of this CreateExperimentalWorkspace.  # noqa: E501

        Id of the associated job.  # noqa: E501

        :return: The cloned_job_id of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._cloned_job_id

    @cloned_job_id.setter
    def cloned_job_id(self, cloned_job_id):
        """Sets the cloned_job_id of this CreateExperimentalWorkspace.

        Id of the associated job.  # noqa: E501

        :param cloned_job_id: The cloned_job_id of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """

        self._cloned_job_id = cloned_job_id

    @property
    def cloned_workspace_id(self):
        """Gets the cloned_workspace_id of this CreateExperimentalWorkspace.  # noqa: E501

        Id of the cloned workspace.  # noqa: E501

        :return: The cloned_workspace_id of this CreateExperimentalWorkspace.  # noqa: E501
        :rtype: str
        """
        return self._cloned_workspace_id

    @cloned_workspace_id.setter
    def cloned_workspace_id(self, cloned_workspace_id):
        """Sets the cloned_workspace_id of this CreateExperimentalWorkspace.

        Id of the cloned workspace.  # noqa: E501

        :param cloned_workspace_id: The cloned_workspace_id of this CreateExperimentalWorkspace.  # noqa: E501
        :type: str
        """

        self._cloned_workspace_id = cloned_workspace_id

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, CreateExperimentalWorkspace):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, CreateExperimentalWorkspace):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ServiceUsage(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'product_type': 'ProductType',
        'credit_card_information': 'CreditCardInformation',
        'bank_account_information': 'BankAccountInformation',
        'current_billing_period_start': 'datetime',
        'current_billing_period_end': 'datetime',
        'unpaid_balance': 'int',
        'available_credits': 'int',
        'current_period_usage': 'int',
        'projected_balance_due': 'int'
    }

    attribute_map = {
        'product_type': 'product_type',
        'credit_card_information': 'credit_card_information',
        'bank_account_information': 'bank_account_information',
        'current_billing_period_start': 'current_billing_period_start',
        'current_billing_period_end': 'current_billing_period_end',
        'unpaid_balance': 'unpaid_balance',
        'available_credits': 'available_credits',
        'current_period_usage': 'current_period_usage',
        'projected_balance_due': 'projected_balance_due'
    }

    def __init__(self, product_type=None, credit_card_information=None, bank_account_information=None, current_billing_period_start=None, current_billing_period_end=None, unpaid_balance=None, available_credits=None, current_period_usage=None, projected_balance_due=None, local_vars_configuration=None):  # noqa: E501
        """ServiceUsage - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._product_type = None
        self._credit_card_information = None
        self._bank_account_information = None
        self._current_billing_period_start = None
        self._current_billing_period_end = None
        self._unpaid_balance = None
        self._available_credits = None
        self._current_period_usage = None
        self._projected_balance_due = None
        self.discriminator = None

        self.product_type = product_type
        if credit_card_information is not None:
            self.credit_card_information = credit_card_information
        if bank_account_information is not None:
            self.bank_account_information = bank_account_information
        self.current_billing_period_start = current_billing_period_start
        self.current_billing_period_end = current_billing_period_end
        self.unpaid_balance = unpaid_balance
        self.available_credits = available_credits
        self.current_period_usage = current_period_usage
        self.projected_balance_due = projected_balance_due

    @property
    def product_type(self):
        """Gets the product_type of this ServiceUsage.  # noqa: E501

        Type of product subscription this information is about.  # noqa: E501

        :return: The product_type of this ServiceUsage.  # noqa: E501
        :rtype: ProductType
        """
        return self._product_type

    @product_type.setter
    def product_type(self, product_type):
        """Sets the product_type of this ServiceUsage.

        Type of product subscription this information is about.  # noqa: E501

        :param product_type: The product_type of this ServiceUsage.  # noqa: E501
        :type: ProductType
        """
        if self.local_vars_configuration.client_side_validation and product_type is None:  # noqa: E501
            raise ValueError("Invalid value for `product_type`, must not be `None`")  # noqa: E501

        self._product_type = product_type

    @property
    def credit_card_information(self):
        """Gets the credit_card_information of this ServiceUsage.  # noqa: E501

        Payment information if payment type is credit card.  # noqa: E501

        :return: The credit_card_information of this ServiceUsage.  # noqa: E501
        :rtype: CreditCardInformation
        """
        return self._credit_card_information

    @credit_card_information.setter
    def credit_card_information(self, credit_card_information):
        """Sets the credit_card_information of this ServiceUsage.

        Payment information if payment type is credit card.  # noqa: E501

        :param credit_card_information: The credit_card_information of this ServiceUsage.  # noqa: E501
        :type: CreditCardInformation
        """

        self._credit_card_information = credit_card_information

    @property
    def bank_account_information(self):
        """Gets the bank_account_information of this ServiceUsage.  # noqa: E501

        Payment information if payment type is bank account.  # noqa: E501

        :return: The bank_account_information of this ServiceUsage.  # noqa: E501
        :rtype: BankAccountInformation
        """
        return self._bank_account_information

    @bank_account_information.setter
    def bank_account_information(self, bank_account_information):
        """Sets the bank_account_information of this ServiceUsage.

        Payment information if payment type is bank account.  # noqa: E501

        :param bank_account_information: The bank_account_information of this ServiceUsage.  # noqa: E501
        :type: BankAccountInformation
        """

        self._bank_account_information = bank_account_information

    @property
    def current_billing_period_start(self):
        """Gets the current_billing_period_start of this ServiceUsage.  # noqa: E501

        Start date for current billing period.  # noqa: E501

        :return: The current_billing_period_start of this ServiceUsage.  # noqa: E501
        :rtype: datetime
        """
        return self._current_billing_period_start

    @current_billing_period_start.setter
    def current_billing_period_start(self, current_billing_period_start):
        """Sets the current_billing_period_start of this ServiceUsage.

        Start date for current billing period.  # noqa: E501

        :param current_billing_period_start: The current_billing_period_start of this ServiceUsage.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and current_billing_period_start is None:  # noqa: E501
            raise ValueError("Invalid value for `current_billing_period_start`, must not be `None`")  # noqa: E501

        self._current_billing_period_start = current_billing_period_start

    @property
    def current_billing_period_end(self):
        """Gets the current_billing_period_end of this ServiceUsage.  # noqa: E501

        End date for current billing period.  # noqa: E501

        :return: The current_billing_period_end of this ServiceUsage.  # noqa: E501
        :rtype: datetime
        """
        return self._current_billing_period_end

    @current_billing_period_end.setter
    def current_billing_period_end(self, current_billing_period_end):
        """Sets the current_billing_period_end of this ServiceUsage.

        End date for current billing period.  # noqa: E501

        :param current_billing_period_end: The current_billing_period_end of this ServiceUsage.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and current_billing_period_end is None:  # noqa: E501
            raise ValueError("Invalid value for `current_billing_period_end`, must not be `None`")  # noqa: E501

        self._current_billing_period_end = current_billing_period_end

    @property
    def unpaid_balance(self):
        """Gets the unpaid_balance of this ServiceUsage.  # noqa: E501

        Accumulated unpaid balances from previous billing periods in cents.  # noqa: E501

        :return: The unpaid_balance of this ServiceUsage.  # noqa: E501
        :rtype: int
        """
        return self._unpaid_balance

    @unpaid_balance.setter
    def unpaid_balance(self, unpaid_balance):
        """Sets the unpaid_balance of this ServiceUsage.

        Accumulated unpaid balances from previous billing periods in cents.  # noqa: E501

        :param unpaid_balance: The unpaid_balance of this ServiceUsage.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and unpaid_balance is None:  # noqa: E501
            raise ValueError("Invalid value for `unpaid_balance`, must not be `None`")  # noqa: E501

        self._unpaid_balance = unpaid_balance

    @property
    def available_credits(self):
        """Gets the available_credits of this ServiceUsage.  # noqa: E501

        Available free and prepaid credits (in cents) at start of current billing cycle (not projected).  # noqa: E501

        :return: The available_credits of this ServiceUsage.  # noqa: E501
        :rtype: int
        """
        return self._available_credits

    @available_credits.setter
    def available_credits(self, available_credits):
        """Sets the available_credits of this ServiceUsage.

        Available free and prepaid credits (in cents) at start of current billing cycle (not projected).  # noqa: E501

        :param available_credits: The available_credits of this ServiceUsage.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and available_credits is None:  # noqa: E501
            raise ValueError("Invalid value for `available_credits`, must not be `None`")  # noqa: E501

        self._available_credits = available_credits

    @property
    def current_period_usage(self):
        """Gets the current_period_usage of this ServiceUsage.  # noqa: E501

        Cost of usage in current billing period (in cents).  # noqa: E501

        :return: The current_period_usage of this ServiceUsage.  # noqa: E501
        :rtype: int
        """
        return self._current_period_usage

    @current_period_usage.setter
    def current_period_usage(self, current_period_usage):
        """Sets the current_period_usage of this ServiceUsage.

        Cost of usage in current billing period (in cents).  # noqa: E501

        :param current_period_usage: The current_period_usage of this ServiceUsage.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and current_period_usage is None:  # noqa: E501
            raise ValueError("Invalid value for `current_period_usage`, must not be `None`")  # noqa: E501

        self._current_period_usage = current_period_usage

    @property
    def projected_balance_due(self):
        """Gets the projected_balance_due of this ServiceUsage.  # noqa: E501

        Projected balance due at end of billing period (for only current billing period).  # noqa: E501

        :return: The projected_balance_due of this ServiceUsage.  # noqa: E501
        :rtype: int
        """
        return self._projected_balance_due

    @projected_balance_due.setter
    def projected_balance_due(self, projected_balance_due):
        """Sets the projected_balance_due of this ServiceUsage.

        Projected balance due at end of billing period (for only current billing period).  # noqa: E501

        :param projected_balance_due: The projected_balance_due of this ServiceUsage.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and projected_balance_due is None:  # noqa: E501
            raise ValueError("Invalid value for `projected_balance_due`, must not be `None`")  # noqa: E501

        self._projected_balance_due = projected_balance_due

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ServiceUsage):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ServiceUsage):
            return True

        return self.to_dict() != other.to_dict()



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class ResourceHistoricalCosts(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'cpu_core_cost': 'HistoricalCosts',
        'memory_gb_cost': 'HistoricalCosts',
        'a100_core_cost': 'HistoricalCosts',
        'k80_core_cost': 'HistoricalCosts',
        'm60_core_cost': 'HistoricalCosts',
        't4_core_cost': 'HistoricalCosts',
        'v100_core_cost': 'HistoricalCosts',
        'v520_core_cost': 'HistoricalCosts'
    }

    attribute_map = {
        'cpu_core_cost': 'cpu_core_cost',
        'memory_gb_cost': 'memory_gb_cost',
        'a100_core_cost': 'a100_core_cost',
        'k80_core_cost': 'k80_core_cost',
        'm60_core_cost': 'm60_core_cost',
        't4_core_cost': 't4_core_cost',
        'v100_core_cost': 'v100_core_cost',
        'v520_core_cost': 'v520_core_cost'
    }

    def __init__(self, cpu_core_cost=None, memory_gb_cost=None, a100_core_cost=None, k80_core_cost=None, m60_core_cost=None, t4_core_cost=None, v100_core_cost=None, v520_core_cost=None, local_vars_configuration=None):  # noqa: E501
        """ResourceHistoricalCosts - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._cpu_core_cost = None
        self._memory_gb_cost = None
        self._a100_core_cost = None
        self._k80_core_cost = None
        self._m60_core_cost = None
        self._t4_core_cost = None
        self._v100_core_cost = None
        self._v520_core_cost = None
        self.discriminator = None

        self.cpu_core_cost = cpu_core_cost
        self.memory_gb_cost = memory_gb_cost
        self.a100_core_cost = a100_core_cost
        self.k80_core_cost = k80_core_cost
        self.m60_core_cost = m60_core_cost
        self.t4_core_cost = t4_core_cost
        self.v100_core_cost = v100_core_cost
        self.v520_core_cost = v520_core_cost

    @property
    def cpu_core_cost(self):
        """Gets the cpu_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on CPU cores by this organization.  # noqa: E501

        :return: The cpu_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._cpu_core_cost

    @cpu_core_cost.setter
    def cpu_core_cost(self, cpu_core_cost):
        """Sets the cpu_core_cost of this ResourceHistoricalCosts.

        The money spent on CPU cores by this organization.  # noqa: E501

        :param cpu_core_cost: The cpu_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and cpu_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `cpu_core_cost`, must not be `None`")  # noqa: E501

        self._cpu_core_cost = cpu_core_cost

    @property
    def memory_gb_cost(self):
        """Gets the memory_gb_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on memory by this organization.  # noqa: E501

        :return: The memory_gb_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._memory_gb_cost

    @memory_gb_cost.setter
    def memory_gb_cost(self, memory_gb_cost):
        """Sets the memory_gb_cost of this ResourceHistoricalCosts.

        The money spent on memory by this organization.  # noqa: E501

        :param memory_gb_cost: The memory_gb_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and memory_gb_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `memory_gb_cost`, must not be `None`")  # noqa: E501

        self._memory_gb_cost = memory_gb_cost

    @property
    def a100_core_cost(self):
        """Gets the a100_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on A100 cores by this organization.  # noqa: E501

        :return: The a100_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._a100_core_cost

    @a100_core_cost.setter
    def a100_core_cost(self, a100_core_cost):
        """Sets the a100_core_cost of this ResourceHistoricalCosts.

        The money spent on A100 cores by this organization.  # noqa: E501

        :param a100_core_cost: The a100_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and a100_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `a100_core_cost`, must not be `None`")  # noqa: E501

        self._a100_core_cost = a100_core_cost

    @property
    def k80_core_cost(self):
        """Gets the k80_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on K80 cores by this organization.  # noqa: E501

        :return: The k80_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._k80_core_cost

    @k80_core_cost.setter
    def k80_core_cost(self, k80_core_cost):
        """Sets the k80_core_cost of this ResourceHistoricalCosts.

        The money spent on K80 cores by this organization.  # noqa: E501

        :param k80_core_cost: The k80_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and k80_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `k80_core_cost`, must not be `None`")  # noqa: E501

        self._k80_core_cost = k80_core_cost

    @property
    def m60_core_cost(self):
        """Gets the m60_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on M60 cores by this organization.  # noqa: E501

        :return: The m60_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._m60_core_cost

    @m60_core_cost.setter
    def m60_core_cost(self, m60_core_cost):
        """Sets the m60_core_cost of this ResourceHistoricalCosts.

        The money spent on M60 cores by this organization.  # noqa: E501

        :param m60_core_cost: The m60_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and m60_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `m60_core_cost`, must not be `None`")  # noqa: E501

        self._m60_core_cost = m60_core_cost

    @property
    def t4_core_cost(self):
        """Gets the t4_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on T4 cores by this organization.  # noqa: E501

        :return: The t4_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._t4_core_cost

    @t4_core_cost.setter
    def t4_core_cost(self, t4_core_cost):
        """Sets the t4_core_cost of this ResourceHistoricalCosts.

        The money spent on T4 cores by this organization.  # noqa: E501

        :param t4_core_cost: The t4_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and t4_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `t4_core_cost`, must not be `None`")  # noqa: E501

        self._t4_core_cost = t4_core_cost

    @property
    def v100_core_cost(self):
        """Gets the v100_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on V100 cores by this organization.  # noqa: E501

        :return: The v100_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._v100_core_cost

    @v100_core_cost.setter
    def v100_core_cost(self, v100_core_cost):
        """Sets the v100_core_cost of this ResourceHistoricalCosts.

        The money spent on V100 cores by this organization.  # noqa: E501

        :param v100_core_cost: The v100_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and v100_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `v100_core_cost`, must not be `None`")  # noqa: E501

        self._v100_core_cost = v100_core_cost

    @property
    def v520_core_cost(self):
        """Gets the v520_core_cost of this ResourceHistoricalCosts.  # noqa: E501

        The money spent on V520 cores by this organization.  # noqa: E501

        :return: The v520_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :rtype: HistoricalCosts
        """
        return self._v520_core_cost

    @v520_core_cost.setter
    def v520_core_cost(self, v520_core_cost):
        """Sets the v520_core_cost of this ResourceHistoricalCosts.

        The money spent on V520 cores by this organization.  # noqa: E501

        :param v520_core_cost: The v520_core_cost of this ResourceHistoricalCosts.  # noqa: E501
        :type: HistoricalCosts
        """
        if self.local_vars_configuration.client_side_validation and v520_core_cost is None:  # noqa: E501
            raise ValueError("Invalid value for `v520_core_cost`, must not be `None`")  # noqa: E501

        self._v520_core_cost = v520_core_cost

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, ResourceHistoricalCosts):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, ResourceHistoricalCosts):
            return True

        return self.to_dict() != other.to_dict()



"""
Fetches data required and formats output for `anyscale cloud` commands.
"""

import ipaddress
import json
import secrets
from typing import Any, Dict, Optional

import boto3
import botocore
from click import ClickException, INT, prompt

from anyscale.aws_iam_policies import (
    AMAZON_ECR_READONLY_ACCESS_POLICY_ARN,
    AMAZON_S3_FULL_ACCESS_POLICY_ARN,
    ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN,
    ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE,
    ANYSCALE_SSM_READ_WRITE_ACCESS_POLICY_DOCUMENT,
    ANYSCALE_SSM_READONLY_ACCESS_POLICY_DOCUMENT,
)
from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client.models import Cloud, CloudConfig, WriteCloud
from anyscale.cloud import get_cloud_id_and_name, get_cloud_json_from_id
from anyscale.conf import ANYSCALE_IAM_ROLE_NAME
from anyscale.controllers.base_controller import BaseController
from anyscale.formatters import clouds_formatter
from anyscale.util import (  # pylint:disable=private-import
    _CACHED_GCP_REGIONS,
    _client,
    _get_role,
    _resource,
    _update_external_ids_for_policy,
    confirm,
    get_available_regions,
    launch_gcp_cloud_setup,
)


ROLE_CREATION_RETRIES = 30
ROLE_CREATION_INTERVAL_SECONDS = 1


class CloudController(BaseController):
    def __init__(
        self, log: BlockLogger = BlockLogger(), initialize_auth_api_client: bool = True
    ):
        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.log.open_block("Output")

    def delete_cloud(
        self,
        cloud_name: Optional[str],
        cloud_id: Optional[str],
        skip_confirmation: bool,
    ) -> bool:
        """
        Deletes a cloud by name or id.
        """

        if not cloud_id and not cloud_name:
            raise ClickException("Must either provide the cloud name or cloud id.")

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )

        confirm(
            f"You'll lose access to existing sessions created with cloud {cloud_id} if you drop it.\nContinue?",
            skip_confirmation,
        )

        self.api_client.delete_cloud_api_v2_clouds_cloud_id_delete(cloud_id=cloud_id)

        self.log.info(f"Deleted cloud {cloud_name}")

        return True

    def list_clouds(self, cloud_name: Optional[str], cloud_id: Optional[str]) -> str:
        if cloud_id is not None:
            clouds = [
                self.api_client.get_cloud_api_v2_clouds_cloud_id_get(cloud_id).result
            ]
        elif cloud_name is not None:
            clouds = [
                self.api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post(
                    {"name": cloud_name}
                ).result
            ]
        else:
            clouds = self.api_client.list_clouds_api_v2_clouds_get().results
        output = clouds_formatter.format_clouds_output(clouds=clouds, json_format=False)

        return str(output)

    def verify_vpc_peering(
        self,
        yes: bool,
        vpc_peering_ip_range: Optional[str],
        vpc_peering_target_project_id: Optional[str],
        vpc_peering_target_vpc_id: Optional[str],
    ) -> None:
        if (
            vpc_peering_ip_range
            or vpc_peering_target_project_id
            or vpc_peering_target_vpc_id
        ):
            if not vpc_peering_ip_range:
                raise ClickException("Please specify a VPC peering IP range.")
            if not vpc_peering_target_project_id:
                raise ClickException("Please specify a VPC peering target project ID.")
            if not vpc_peering_target_vpc_id:
                raise ClickException("Please specify a VPC peering target VPC ID.")
        else:
            return

        try:
            valid_ip_network = ipaddress.IPv4Network(vpc_peering_ip_range)
        except ValueError:
            raise ClickException(f"{vpc_peering_ip_range} is not a valid IP address.")
        # https://cloud.google.com/vpc/docs/vpc#valid-ranges
        allowed_ip_ranges = [
            ipaddress.IPv4Network("10.0.0.0/8"),
            ipaddress.IPv4Network("172.16.0.0/12"),
            ipaddress.IPv4Network("192.168.0.0/16"),
        ]

        for allowed_ip_range in allowed_ip_ranges:
            if valid_ip_network.subnet_of(allowed_ip_range):
                break
        else:
            raise ClickException(
                f"{vpc_peering_ip_range} is not a allowed private IP address range for GCP. The allowed IP ranges are 10.0.0.0/8, 172.16.0.0/12, and 192.168.0.0/16. For more info, see https://cloud.google.com/vpc/docs/vpc#valid-ranges"
            )

        if (
            valid_ip_network.num_addresses
            < ipaddress.IPv4Network("192.168.0.0/16").num_addresses
        ):
            raise ClickException(
                f"{vpc_peering_ip_range} is not a valid IP range. The minimum size is /16"
            )

        if not yes:
            confirm(
                f"\nYou selected to create a VPC peering connection to VPC {vpc_peering_target_vpc_id} in GCP project {vpc_peering_target_project_id}."
                f"This will create a VPC peering connection from your Anyscale GCP project to the target project ({vpc_peering_target_project_id})."
                "You will need to manually create the peering connection from the target project to your Anyscale GCP project after the anyscale cloud is created.\n"
                "Continue cloud setup?",
                False,
            )

    def setup_cloud(
        self,
        provider: str,
        region: Optional[str],
        name: str,
        yes: bool = False,
        folder_id: Optional[int] = None,
        vpc_peering_ip_range: Optional[str] = None,
        vpc_peering_target_project_id: Optional[str] = None,
        vpc_peering_target_vpc_id: Optional[str] = None,
    ) -> None:
        """
        Sets up a cloud provider
        """
        if provider == "aws":
            # If the region is blank, change it to the default for AWS.
            if region is None:
                region = "us-west-2"
            regions_available = get_available_regions()
            if region not in regions_available:
                raise ClickException(
                    f"Region '{region}' is not available. Regions available are: "
                    f"{', '.join(map(repr, regions_available))}"
                )
            self.setup_aws(region=region, name=name, yes=yes)
        elif provider == "gcp":
            # If the region is blank, change it to the default for GCP.
            if region is None:
                region = "us-west1"
            # Warn the user about a bad region before the cloud configuration begins.
            # GCP's `list regions` API requires a project, meaning true verification
            # happens in the middle of the flow.
            if region not in _CACHED_GCP_REGIONS and not yes:
                confirm(
                    f"You selected the region: {region}, but it is not in"
                    f"the cached list of GCP regions:\n\n{_CACHED_GCP_REGIONS}.\n"
                    "Continue cloud setup with this region?",
                    False,
                )
            if not yes and not folder_id:
                folder_id = prompt(
                    "Please select the GCP Folder ID where the 'Anyscale' folder will be created.\n"
                    "\tYour GCP account must have permissions to create sub-folders in the specified folder.\n"
                    "\tView your organization's folder layout here: https://console.cloud.google.com/cloud-resource-manager\n"
                    "\tIf not specified, the 'Anyscale' folder will be created directly under the organization.\n"
                    "Folder ID (numerals only)",
                    default="",
                    type=INT,
                    show_default=False,
                )

            self.verify_vpc_peering(
                yes,
                vpc_peering_ip_range,
                vpc_peering_target_project_id,
                vpc_peering_target_vpc_id,
            )
            # TODO: interactive setup process through the CLI?
            launch_gcp_cloud_setup(
                name=name,
                region=region,
                folder_id=folder_id,
                vpc_peering_ip_range=vpc_peering_ip_range,
                vpc_peering_target_project_id=vpc_peering_target_project_id,
                vpc_peering_target_vpc_id=vpc_peering_target_vpc_id,
            )
        else:
            raise ClickException(
                f"Invalid Cloud provider: {provider}. Available providers are [aws, gcp]."
            )

    def setup_aws(self, region: str, name: str, yes: bool = False) -> None:
        from ray.autoscaler._private.aws.config import DEFAULT_RAY_IAM_ROLE

        confirm(
            "\nYou are about to give Anyscale access to EC2 permissions necessary to manage clusters.\n"
            "A separate AWS role is created for your clusters to run with \nand will be granted readonly access to ECR & S3 Full Access in your AWS account.\n\n"
            "Continue?",
            yes,
        )

        self.setup_aws_cross_account_role(region, name)
        self.setup_aws_ray_role(region, DEFAULT_RAY_IAM_ROLE)

        self.log.info("AWS credentials setup complete.")
        self.log.info(
            "You can revoke the access at any time by deleting anyscale IAM user/role in your account."
        )
        self.log.info(
            "Head over to the web UI to create new sessions in your AWS account."
        )

    def setup_aws_cross_account_role(self, region: str, name: str) -> None:
        response = (
            self.api_client.get_anyscale_aws_account_api_v2_clouds_anyscale_aws_account_get()
        )

        anyscale_aws_account = response.result.anyscale_aws_account
        anyscale_aws_iam_role_policy: Dict[str, Any] = {
            "Version": "2012-10-17",
            "Statement": {
                "Sid": "1",
                "Effect": "Allow",
                "Principal": {"AWS": anyscale_aws_account},
                "Action": "sts:AssumeRole",
            },
        }

        for _ in range(5):
            role_name = "{}-{}".format(ANYSCALE_IAM_ROLE_NAME, secrets.token_hex(4))

            role = _get_role(role_name, region)
            if role is None:
                break
        else:
            raise RuntimeError(
                "We weren't able to connect your account with the Anyscale because we weren't able to find an available IAM Role name in your account. Please reach out to support or your SA for assistance."
            )

        iam = _resource("iam", region)
        iam.create_role(
            RoleName=role_name,
            AssumeRolePolicyDocument=json.dumps(anyscale_aws_iam_role_policy),
        )
        role = _get_role(role_name, region)

        assert role is not None, "Failed to create IAM role."

        role.Policy(name="Anyscale_IAM_Policy_Steady_State").put(
            PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_STEADY_STATE)
        )

        role.Policy(name="Anyscale_IAM_Policy_Initial_Setup").put(
            PolicyDocument=json.dumps(ANYSCALE_IAM_PERMISSIONS_EC2_INITIAL_RUN)
        )

        self.log.info(f"Using IAM role {role.arn}")
        try:

            created_cloud = self.api_client.create_cloud_api_v2_clouds_post(
                write_cloud=WriteCloud(
                    provider="AWS", region=region, credentials=role.arn, name=name,
                )
            )
        except ClickException:
            self.log.info("Create failed, cleaning up IAM role: {}".format(role_name))
            try:
                for policy in role.policies.all():
                    policy.delete()
                role.delete()
            except botocore.exceptions.ClientError:
                self.log.error(
                    "Failed to clean up IAM role after a failed cloud creation: {}".format(
                        role_name
                    )
                )
            raise
        cloud_id = created_cloud.result.id

        iam_client = _client("iam", region)
        iam_client.update_role(
            RoleName=role.name,
            Description="Anyscale access role for cloud {} in region {}".format(
                cloud_id, created_cloud.result.region
            ),
        )

        # NOTE: We update this _after_ cloud creation because this External ID MUST
        # come from Anyscale, not the customer. We are using the `cloud_id` as it is unique per cloud.
        # https://docs.aws.amazon.com/IAM/latest/UserGuide/id_roles_create_for-user_externalid.html
        new_policy = _update_external_ids_for_policy(
            role.assume_role_policy_document, cloud_id
        )

        role.AssumeRolePolicy().update(PolicyDocument=json.dumps(new_policy))

    def setup_aws_ray_role(self, region: str, role_name: str) -> None:
        iam = boto3.resource("iam", region_name=region)

        role = _get_role(role_name, region)
        if role is None:
            iam.create_role(
                RoleName=role_name,
                AssumeRolePolicyDocument=json.dumps(
                    {
                        "Version": "2012-10-17",
                        "Statement": {
                            "Effect": "Allow",
                            "Principal": {"Service": ["ec2.amazonaws.com"]},
                            "Action": "sts:AssumeRole",
                        },
                    }
                ),
            )

            role = _get_role(role_name, region)

        role.attach_policy(PolicyArn=AMAZON_ECR_READONLY_ACCESS_POLICY_ARN)
        # Modified permissions from Ray (no EC2FullAccess)
        role.attach_policy(PolicyArn=AMAZON_S3_FULL_ACCESS_POLICY_ARN)

        for profile in role.instance_profiles.all():
            if profile.name == role_name:
                return
        profile = iam.create_instance_profile(InstanceProfileName=role_name)
        profile.add_role(RoleName=role_name)

    def update_cloud_config(
        self,
        cloud_name: Optional[str],
        cloud_id: Optional[str],
        max_stopped_instances: int,
    ) -> None:
        """Updates a cloud's configuration by name or id.

        Currently the only supported option is "max_stopped_instances."
        """

        if not cloud_id and not cloud_name:
            raise ClickException("Must either provide the cloud name or cloud id.")

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )

        self.api_client.update_cloud_config_api_v2_clouds_cloud_id_config_put(
            cloud_id=cloud_id,
            cloud_config=CloudConfig(max_stopped_instances=max_stopped_instances),
        )

        self.log.info(f"Updated config for cloud '{cloud_name}' to:")
        self.log.info(self.get_cloud_config(cloud_name=None, cloud_id=cloud_id))

    def get_cloud_config(
        self, cloud_name: Optional[str] = None, cloud_id: Optional[str] = None,
    ) -> str:
        """Get a cloud's current JSON configuration."""

        if not cloud_id and not cloud_name:
            raise ClickException("Must either provide the cloud name or cloud id.")

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )

        return str(get_cloud_json_from_id(cloud_id, self.api_client)["config"])

    def set_default_cloud(
        self, cloud_name: Optional[str], cloud_id: Optional[str],
    ) -> None:
        """
        Sets default cloud for caller's organization. This operation can only be performed
        by organization admins, and the default cloud must have organization level
        permissions.
        """

        if not cloud_id and not cloud_name:
            raise ClickException("Must either provide the cloud name or cloud id.")

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )

        self.api_client.update_default_cloud_api_v2_organizations_update_default_cloud_post(
            cloud_id=cloud_id
        )

        self.log.info(f"Updated default cloud to {cloud_name}")

    def experimental_setup_secrets(
        self,
        cloud_name: Optional[str],
        cloud_id: Optional[str],
        write_permissions: bool,
        yes: bool,
    ):
        """
        Given a cloud name, look up its provider and give it permissions to read secrets
        """
        feature_flag_on = self.api_client.check_is_feature_flag_on_api_v2_userinfo_check_is_feature_flag_on_get(
            "wandb-integration-prototype"
        ).result.is_on
        if not feature_flag_on:
            raise ClickException(
                "Secrets can only be set up if the feature flag is enabled. "
                "Please contact Anyscale support to enable the flag."
            )

        cloud_id, cloud_name = get_cloud_id_and_name(
            self.api_client, cloud_id, cloud_name
        )
        cloud = self.api_client.get_cloud_api_v2_clouds_cloud_id_get(cloud_id).result

        self.log.info(
            f"Setting up secrets policy for {cloud.provider} cloud {cloud.name}"
        )

        if cloud.provider == "AWS":
            return self._experimental_grant_secrets_access_aws(
                cloud, write_permissions, yes
            )

        if cloud.provider == "GCP":
            raise ClickException(
                "Cloud secrets not currently supported for GCP clouds."
            )

        raise ClickException(f"Cloud secrets not supported for {cloud_name}")

    def _experimental_grant_secrets_access_aws(
        self, cloud: Cloud, write_permissions: bool, yes: bool
    ) -> None:
        """Creates IAM policy for SSM readonly access and attaches it to the given role
        Args:
            cloud (Cloud): Cloud object which needs modification
            write_permissions (bool): Whether to add write permissions for Secrets Manager
                to policy
        """

        # Ensure they are in the correct AWS account, by checking account used
        # for Security Token Service
        STS = "sts"
        current_account = boto3.client(
            STS, region_name=cloud.region
        ).get_caller_identity()["Account"]
        # Cloud credentials of format arn:aws:iam::{cloud_account}:role/{cloud_role}
        # Split credentials to get cloud account.
        cloud_account = cloud.credentials.split(":")[4]

        if current_account != cloud_account:
            raise ClickException(
                f"The cloud you specified uses AWS account {cloud_account}, "
                f"but you are currently logged into {current_account}."
            )

        from ray.autoscaler._private.aws.config import DEFAULT_RAY_IAM_ROLE

        if yes:
            role_name = DEFAULT_RAY_IAM_ROLE
        else:
            role_name = prompt(
                "Which AWS role do you want to grant readonly SSM access to?",
                default=DEFAULT_RAY_IAM_ROLE,
                show_default=True,
            )

        role = _get_role(role_name, cloud.region)
        assert (
            role is not None
        ), f"Failed to find IAM role {role_name} in Cloud {cloud.name}! Have you run 'cloud setup'?"

        policy_name = (
            f"anyscale-secrets-read-write-{cloud.id}"
            if write_permissions
            else f"anyscale-secrets-readonly-{cloud.id}"
        )
        policy_document = (
            ANYSCALE_SSM_READ_WRITE_ACCESS_POLICY_DOCUMENT
            if write_permissions
            else ANYSCALE_SSM_READONLY_ACCESS_POLICY_DOCUMENT
        )

        role.Policy(name=policy_name).put(PolicyDocument=json.dumps(policy_document),)
        self.log.info(
            f"Successfully added/updated inline policy {policy_name} on role {role_name}."
        )
        if role_name == DEFAULT_RAY_IAM_ROLE:
            self.log.info(
                f"Note: {role_name} is the default role used for all Anyscale clouds in "
                f"this AWS account, so policy {policy_name} will be used by all clouds "
                "that use this role. We are planning to create a new default role for each "
                "Anyscale cloud in the future."
            )



# AUTOGENERATED - modify shared_anyscale_util in root directory to make changes
import base64
from dataclasses import asdict, dataclass
import urllib.parse


BYOD_PREFIX = "custom_docker_"
BYOD_DISPLAY_PREFIX = "docker://"


def is_byod_id(id: str) -> bool:
    return id.startswith(BYOD_PREFIX)


@dataclass(frozen=True)
class BYODInfo:
    docker_image_name: str
    python_version: str
    ray_version: str

    def encode(self) -> str:
        return BYOD_PREFIX + base64.b64encode(
            urllib.parse.urlencode(asdict(self)).encode("ascii")
        ).decode("ascii")

    def get_name(self) -> str:
        return f"{BYOD_DISPLAY_PREFIX}{self.docker_image_name}"


def decode_byod(id: str) -> BYODInfo:
    if not is_byod_id(id):
        raise ValueError(f"Expected id to start with {BYOD_PREFIX}, got {id}")
    decoded = urllib.parse.parse_qs(
        base64.b64decode(id[len(BYOD_PREFIX) :]).decode("ascii")
    )
    return BYODInfo(
        decoded["docker_image_name"][0],
        decoded["python_version"][0],
        decoded["ray_version"][0],
    )



# coding: utf-8

"""
    Anyscale API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from anyscale_client.configuration import Configuration


class Cluster(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'name': 'str',
        'project_id': 'str',
        'cluster_environment_build_id': 'str',
        'cluster_compute_id': 'str',
        'cluster_compute_config': 'ClusterComputeConfig',
        'idle_timeout_minutes': 'int',
        'allow_public_internet_traffic': 'bool',
        'user_service_access': 'UserServiceAccessTypes',
        'id': 'str',
        'state': 'ClusterState',
        'goal_state': 'ClusterState',
        'creator_id': 'str',
        'created_at': 'datetime',
        'access_token': 'str',
        'services_urls': 'ClusterServicesUrls',
        'head_node_info': 'ClusterHeadNodeInfo',
        'ssh_authorized_keys': 'list[str]',
        'ssh_private_key': 'str',
        'ray_version': 'str',
        'ray_version_last_updated_at': 'datetime',
        'user_service_token': 'str'
    }

    attribute_map = {
        'name': 'name',
        'project_id': 'project_id',
        'cluster_environment_build_id': 'cluster_environment_build_id',
        'cluster_compute_id': 'cluster_compute_id',
        'cluster_compute_config': 'cluster_compute_config',
        'idle_timeout_minutes': 'idle_timeout_minutes',
        'allow_public_internet_traffic': 'allow_public_internet_traffic',
        'user_service_access': 'user_service_access',
        'id': 'id',
        'state': 'state',
        'goal_state': 'goal_state',
        'creator_id': 'creator_id',
        'created_at': 'created_at',
        'access_token': 'access_token',
        'services_urls': 'services_urls',
        'head_node_info': 'head_node_info',
        'ssh_authorized_keys': 'ssh_authorized_keys',
        'ssh_private_key': 'ssh_private_key',
        'ray_version': 'ray_version',
        'ray_version_last_updated_at': 'ray_version_last_updated_at',
        'user_service_token': 'user_service_token'
    }

    def __init__(self, name=None, project_id=None, cluster_environment_build_id=None, cluster_compute_id=None, cluster_compute_config=None, idle_timeout_minutes=120, allow_public_internet_traffic=False, user_service_access=None, id=None, state=None, goal_state=None, creator_id=None, created_at=None, access_token=None, services_urls=None, head_node_info=None, ssh_authorized_keys=None, ssh_private_key=None, ray_version=None, ray_version_last_updated_at=None, user_service_token=None, local_vars_configuration=None):  # noqa: E501
        """Cluster - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._name = None
        self._project_id = None
        self._cluster_environment_build_id = None
        self._cluster_compute_id = None
        self._cluster_compute_config = None
        self._idle_timeout_minutes = None
        self._allow_public_internet_traffic = None
        self._user_service_access = None
        self._id = None
        self._state = None
        self._goal_state = None
        self._creator_id = None
        self._created_at = None
        self._access_token = None
        self._services_urls = None
        self._head_node_info = None
        self._ssh_authorized_keys = None
        self._ssh_private_key = None
        self._ray_version = None
        self._ray_version_last_updated_at = None
        self._user_service_token = None
        self.discriminator = None

        self.name = name
        if project_id is not None:
            self.project_id = project_id
        self.cluster_environment_build_id = cluster_environment_build_id
        self.cluster_compute_id = cluster_compute_id
        if cluster_compute_config is not None:
            self.cluster_compute_config = cluster_compute_config
        if idle_timeout_minutes is not None:
            self.idle_timeout_minutes = idle_timeout_minutes
        if allow_public_internet_traffic is not None:
            self.allow_public_internet_traffic = allow_public_internet_traffic
        if user_service_access is not None:
            self.user_service_access = user_service_access
        self.id = id
        self.state = state
        if goal_state is not None:
            self.goal_state = goal_state
        self.creator_id = creator_id
        self.created_at = created_at
        self.access_token = access_token
        self.services_urls = services_urls
        if head_node_info is not None:
            self.head_node_info = head_node_info
        self.ssh_authorized_keys = ssh_authorized_keys
        self.ssh_private_key = ssh_private_key
        if ray_version is not None:
            self.ray_version = ray_version
        if ray_version_last_updated_at is not None:
            self.ray_version_last_updated_at = ray_version_last_updated_at
        if user_service_token is not None:
            self.user_service_token = user_service_token

    @property
    def name(self):
        """Gets the name of this Cluster.  # noqa: E501

        Name of this Cluster.  # noqa: E501

        :return: The name of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this Cluster.

        Name of this Cluster.  # noqa: E501

        :param name: The name of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and name is None:  # noqa: E501
            raise ValueError("Invalid value for `name`, must not be `None`")  # noqa: E501

        self._name = name

    @property
    def project_id(self):
        """Gets the project_id of this Cluster.  # noqa: E501

        Project that this Cluster belongs to. If none, this Cluster will use the default Project.  # noqa: E501

        :return: The project_id of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._project_id

    @project_id.setter
    def project_id(self, project_id):
        """Sets the project_id of this Cluster.

        Project that this Cluster belongs to. If none, this Cluster will use the default Project.  # noqa: E501

        :param project_id: The project_id of this Cluster.  # noqa: E501
        :type: str
        """

        self._project_id = project_id

    @property
    def cluster_environment_build_id(self):
        """Gets the cluster_environment_build_id of this Cluster.  # noqa: E501

        Cluster Environment Build that this Cluster is using.  # noqa: E501

        :return: The cluster_environment_build_id of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._cluster_environment_build_id

    @cluster_environment_build_id.setter
    def cluster_environment_build_id(self, cluster_environment_build_id):
        """Sets the cluster_environment_build_id of this Cluster.

        Cluster Environment Build that this Cluster is using.  # noqa: E501

        :param cluster_environment_build_id: The cluster_environment_build_id of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_environment_build_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_environment_build_id`, must not be `None`")  # noqa: E501

        self._cluster_environment_build_id = cluster_environment_build_id

    @property
    def cluster_compute_id(self):
        """Gets the cluster_compute_id of this Cluster.  # noqa: E501

        Cluster Compute that this Cluster is using.  # noqa: E501

        :return: The cluster_compute_id of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._cluster_compute_id

    @cluster_compute_id.setter
    def cluster_compute_id(self, cluster_compute_id):
        """Sets the cluster_compute_id of this Cluster.

        Cluster Compute that this Cluster is using.  # noqa: E501

        :param cluster_compute_id: The cluster_compute_id of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and cluster_compute_id is None:  # noqa: E501
            raise ValueError("Invalid value for `cluster_compute_id`, must not be `None`")  # noqa: E501

        self._cluster_compute_id = cluster_compute_id

    @property
    def cluster_compute_config(self):
        """Gets the cluster_compute_config of this Cluster.  # noqa: E501

        One-off cluster compute that this cluster is using.  # noqa: E501

        :return: The cluster_compute_config of this Cluster.  # noqa: E501
        :rtype: ClusterComputeConfig
        """
        return self._cluster_compute_config

    @cluster_compute_config.setter
    def cluster_compute_config(self, cluster_compute_config):
        """Sets the cluster_compute_config of this Cluster.

        One-off cluster compute that this cluster is using.  # noqa: E501

        :param cluster_compute_config: The cluster_compute_config of this Cluster.  # noqa: E501
        :type: ClusterComputeConfig
        """

        self._cluster_compute_config = cluster_compute_config

    @property
    def idle_timeout_minutes(self):
        """Gets the idle_timeout_minutes of this Cluster.  # noqa: E501

        Idle timeout (in minutes), after which the Cluster is terminated. Idle time is defined as the time during which a Cluster is not running a user command (through 'anyscale exec' or the Web UI), and does not have an attached driver. Time spent running Jupyter commands, or commands run through ssh, is still considered 'idle'.  # noqa: E501

        :return: The idle_timeout_minutes of this Cluster.  # noqa: E501
        :rtype: int
        """
        return self._idle_timeout_minutes

    @idle_timeout_minutes.setter
    def idle_timeout_minutes(self, idle_timeout_minutes):
        """Sets the idle_timeout_minutes of this Cluster.

        Idle timeout (in minutes), after which the Cluster is terminated. Idle time is defined as the time during which a Cluster is not running a user command (through 'anyscale exec' or the Web UI), and does not have an attached driver. Time spent running Jupyter commands, or commands run through ssh, is still considered 'idle'.  # noqa: E501

        :param idle_timeout_minutes: The idle_timeout_minutes of this Cluster.  # noqa: E501
        :type: int
        """

        self._idle_timeout_minutes = idle_timeout_minutes

    @property
    def allow_public_internet_traffic(self):
        """Gets the allow_public_internet_traffic of this Cluster.  # noqa: E501

        Whether public internet traffic can access Serve endpoints or if an authentication token is required.  # noqa: E501

        :return: The allow_public_internet_traffic of this Cluster.  # noqa: E501
        :rtype: bool
        """
        return self._allow_public_internet_traffic

    @allow_public_internet_traffic.setter
    def allow_public_internet_traffic(self, allow_public_internet_traffic):
        """Sets the allow_public_internet_traffic of this Cluster.

        Whether public internet traffic can access Serve endpoints or if an authentication token is required.  # noqa: E501

        :param allow_public_internet_traffic: The allow_public_internet_traffic of this Cluster.  # noqa: E501
        :type: bool
        """

        self._allow_public_internet_traffic = allow_public_internet_traffic

    @property
    def user_service_access(self):
        """Gets the user_service_access of this Cluster.  # noqa: E501

        Whether user service can be accessed by public internet traffic.  # noqa: E501

        :return: The user_service_access of this Cluster.  # noqa: E501
        :rtype: UserServiceAccessTypes
        """
        return self._user_service_access

    @user_service_access.setter
    def user_service_access(self, user_service_access):
        """Sets the user_service_access of this Cluster.

        Whether user service can be accessed by public internet traffic.  # noqa: E501

        :param user_service_access: The user_service_access of this Cluster.  # noqa: E501
        :type: UserServiceAccessTypes
        """

        self._user_service_access = user_service_access

    @property
    def id(self):
        """Gets the id of this Cluster.  # noqa: E501

        Server assigned unique identifier.  # noqa: E501

        :return: The id of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this Cluster.

        Server assigned unique identifier.  # noqa: E501

        :param id: The id of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def state(self):
        """Gets the state of this Cluster.  # noqa: E501

        Current state of the Cluster.  # noqa: E501

        :return: The state of this Cluster.  # noqa: E501
        :rtype: ClusterState
        """
        return self._state

    @state.setter
    def state(self, state):
        """Sets the state of this Cluster.

        Current state of the Cluster.  # noqa: E501

        :param state: The state of this Cluster.  # noqa: E501
        :type: ClusterState
        """
        if self.local_vars_configuration.client_side_validation and state is None:  # noqa: E501
            raise ValueError("Invalid value for `state`, must not be `None`")  # noqa: E501

        self._state = state

    @property
    def goal_state(self):
        """Gets the goal_state of this Cluster.  # noqa: E501

        State that this Cluster will eventually transition to. This will not be populated if there are no pending transitions.  # noqa: E501

        :return: The goal_state of this Cluster.  # noqa: E501
        :rtype: ClusterState
        """
        return self._goal_state

    @goal_state.setter
    def goal_state(self, goal_state):
        """Sets the goal_state of this Cluster.

        State that this Cluster will eventually transition to. This will not be populated if there are no pending transitions.  # noqa: E501

        :param goal_state: The goal_state of this Cluster.  # noqa: E501
        :type: ClusterState
        """

        self._goal_state = goal_state

    @property
    def creator_id(self):
        """Gets the creator_id of this Cluster.  # noqa: E501

        User who created this Cluster.  # noqa: E501

        :return: The creator_id of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this Cluster.

        User who created this Cluster.  # noqa: E501

        :param creator_id: The creator_id of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def created_at(self):
        """Gets the created_at of this Cluster.  # noqa: E501

        Time at which this Cluster was created.  # noqa: E501

        :return: The created_at of this Cluster.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this Cluster.

        Time at which this Cluster was created.  # noqa: E501

        :param created_at: The created_at of this Cluster.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def access_token(self):
        """Gets the access_token of this Cluster.  # noqa: E501

        Access token for web based services (e.g. jupyter, tensorboard, etc). This field will be populated when the web based services are available after the Cluster finishes starting.  # noqa: E501

        :return: The access_token of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._access_token

    @access_token.setter
    def access_token(self, access_token):
        """Sets the access_token of this Cluster.

        Access token for web based services (e.g. jupyter, tensorboard, etc). This field will be populated when the web based services are available after the Cluster finishes starting.  # noqa: E501

        :param access_token: The access_token of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and access_token is None:  # noqa: E501
            raise ValueError("Invalid value for `access_token`, must not be `None`")  # noqa: E501

        self._access_token = access_token

    @property
    def services_urls(self):
        """Gets the services_urls of this Cluster.  # noqa: E501

        URLs for additional services running on this Cluster (e.g. Jupyter, Ray Dashboard, etc.).  # noqa: E501

        :return: The services_urls of this Cluster.  # noqa: E501
        :rtype: ClusterServicesUrls
        """
        return self._services_urls

    @services_urls.setter
    def services_urls(self, services_urls):
        """Sets the services_urls of this Cluster.

        URLs for additional services running on this Cluster (e.g. Jupyter, Ray Dashboard, etc.).  # noqa: E501

        :param services_urls: The services_urls of this Cluster.  # noqa: E501
        :type: ClusterServicesUrls
        """
        if self.local_vars_configuration.client_side_validation and services_urls is None:  # noqa: E501
            raise ValueError("Invalid value for `services_urls`, must not be `None`")  # noqa: E501

        self._services_urls = services_urls

    @property
    def head_node_info(self):
        """Gets the head_node_info of this Cluster.  # noqa: E501

        Detailed information about this Cluster's head node. This will only be populated for Clusters that have finished starting.  # noqa: E501

        :return: The head_node_info of this Cluster.  # noqa: E501
        :rtype: ClusterHeadNodeInfo
        """
        return self._head_node_info

    @head_node_info.setter
    def head_node_info(self, head_node_info):
        """Sets the head_node_info of this Cluster.

        Detailed information about this Cluster's head node. This will only be populated for Clusters that have finished starting.  # noqa: E501

        :param head_node_info: The head_node_info of this Cluster.  # noqa: E501
        :type: ClusterHeadNodeInfo
        """

        self._head_node_info = head_node_info

    @property
    def ssh_authorized_keys(self):
        """Gets the ssh_authorized_keys of this Cluster.  # noqa: E501

        Serialized SSH Public Keys to be placed in the machine's authorized_keys.  # noqa: E501

        :return: The ssh_authorized_keys of this Cluster.  # noqa: E501
        :rtype: list[str]
        """
        return self._ssh_authorized_keys

    @ssh_authorized_keys.setter
    def ssh_authorized_keys(self, ssh_authorized_keys):
        """Sets the ssh_authorized_keys of this Cluster.

        Serialized SSH Public Keys to be placed in the machine's authorized_keys.  # noqa: E501

        :param ssh_authorized_keys: The ssh_authorized_keys of this Cluster.  # noqa: E501
        :type: list[str]
        """
        if self.local_vars_configuration.client_side_validation and ssh_authorized_keys is None:  # noqa: E501
            raise ValueError("Invalid value for `ssh_authorized_keys`, must not be `None`")  # noqa: E501

        self._ssh_authorized_keys = ssh_authorized_keys

    @property
    def ssh_private_key(self):
        """Gets the ssh_private_key of this Cluster.  # noqa: E501

        SSH Private key that can be used to access the Cluster's servers.  # noqa: E501

        :return: The ssh_private_key of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._ssh_private_key

    @ssh_private_key.setter
    def ssh_private_key(self, ssh_private_key):
        """Sets the ssh_private_key of this Cluster.

        SSH Private key that can be used to access the Cluster's servers.  # noqa: E501

        :param ssh_private_key: The ssh_private_key of this Cluster.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ssh_private_key is None:  # noqa: E501
            raise ValueError("Invalid value for `ssh_private_key`, must not be `None`")  # noqa: E501

        self._ssh_private_key = ssh_private_key

    @property
    def ray_version(self):
        """Gets the ray_version of this Cluster.  # noqa: E501

        The last known ray version running on this cluster.  # noqa: E501

        :return: The ray_version of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._ray_version

    @ray_version.setter
    def ray_version(self, ray_version):
        """Sets the ray_version of this Cluster.

        The last known ray version running on this cluster.  # noqa: E501

        :param ray_version: The ray_version of this Cluster.  # noqa: E501
        :type: str
        """

        self._ray_version = ray_version

    @property
    def ray_version_last_updated_at(self):
        """Gets the ray_version_last_updated_at of this Cluster.  # noqa: E501

        The time in which the ray version of this cluster was updated.  # noqa: E501

        :return: The ray_version_last_updated_at of this Cluster.  # noqa: E501
        :rtype: datetime
        """
        return self._ray_version_last_updated_at

    @ray_version_last_updated_at.setter
    def ray_version_last_updated_at(self, ray_version_last_updated_at):
        """Sets the ray_version_last_updated_at of this Cluster.

        The time in which the ray version of this cluster was updated.  # noqa: E501

        :param ray_version_last_updated_at: The ray_version_last_updated_at of this Cluster.  # noqa: E501
        :type: datetime
        """

        self._ray_version_last_updated_at = ray_version_last_updated_at

    @property
    def user_service_token(self):
        """Gets the user_service_token of this Cluster.  # noqa: E501

        Token used to authenticate user service if it is accessible to public internet. This field will beempty if user service is not publically accessible.  # noqa: E501

        :return: The user_service_token of this Cluster.  # noqa: E501
        :rtype: str
        """
        return self._user_service_token

    @user_service_token.setter
    def user_service_token(self, user_service_token):
        """Sets the user_service_token of this Cluster.

        Token used to authenticate user service if it is accessible to public internet. This field will beempty if user service is not publically accessible.  # noqa: E501

        :param user_service_token: The user_service_token of this Cluster.  # noqa: E501
        :type: str
        """

        self._user_service_token = user_service_token

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, Cluster):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, Cluster):
            return True

        return self.to_dict() != other.to_dict()



from datetime import datetime
import enum
import os
from typing import Any, Dict, List, Optional

import click
from pydantic import Field
import yaml

from anyscale.cli_logger import BlockLogger
from anyscale.client.openapi_client import (
    ComputeTemplate,
    CreateProductionService,
    ProductionJob,
    ProductionJobConfig,
)
from anyscale.controllers.base_controller import BaseController
from anyscale.controllers.job_controller import JobConfig, JobController
from anyscale.project import infer_project_id
from anyscale.util import get_endpoint, is_anyscale_workspace
from anyscale.utils.runtime_env import override_runtime_env_for_local_working_dir


class UserServiceAccessTypes(str, enum.Enum):
    private = "private"
    public = "public"


class ServiceConfig(JobConfig):
    healthcheck_url: str = Field(..., description="Healthcheck url for service.")
    access: UserServiceAccessTypes = Field(
        UserServiceAccessTypes.public,
        description=(
            "Whether user service (eg: serve deployment) can be accessed by public "
            "internet traffic. If public, a user service endpoint can be queried from "
            "the public internet with the provided authentication token. "
            "If private, the user service endpoint can only be queried from within "
            "the same Anyscale cloud and will not require an authentication token."
        ),
    )


class ServiceController(BaseController):
    def __init__(
        self, log: BlockLogger = BlockLogger(), initialize_auth_api_client: bool = True
    ):
        super().__init__(initialize_auth_api_client=initialize_auth_api_client)
        self.log = log
        self.job_controller = JobController(
            initialize_auth_api_client=initialize_auth_api_client
        )

    def deploy(
        self,
        service_config_file: str,
        name: Optional[str],
        description: Optional[str],
        healthcheck_url: Optional[str] = None,
        is_entrypoint_cmd: Optional[bool] = False,
        entrypoint: Optional[List[str]] = None,
    ) -> None:
        entrypoint = entrypoint or []
        if is_anyscale_workspace() and is_entrypoint_cmd:
            entrypoint = [service_config_file, *entrypoint]
            config = self.generate_config_from_entrypoint(
                entrypoint, name, description, healthcheck_url=healthcheck_url
            )
            self.deploy_from_config(config)
        elif len(entrypoint) == 0:
            # Assume that job_config_file is a file and submit it.
            config = self.generate_config_from_file(
                service_config_file, name, description, healthcheck_url=healthcheck_url,
            )
            self.deploy_from_config(config)
        elif len(entrypoint) != 0:
            msg = (
                "Within an Anyscale Workspace, `anyscale service deploy` takes either a file, or a command. To submit a command, use `anyscale service deploy -- my command`."
                if is_anyscale_workspace()
                else "`anyscale service deploy` takes one argument, a YAML file configuration. Please use `anyscale service deploy my_file`."
            )
            raise click.ClickException(msg)

    def generate_config_from_entrypoint(
        self,
        entrypoint: List[str],
        name: Optional[str],
        description: Optional[str],
        healthcheck_url: Optional[str] = None,
    ) -> ServiceConfig:
        config_dict = {
            "entrypoint": " ".join(entrypoint),
            "name": name,
            "description": description,
            "healthcheck_url": healthcheck_url,
        }
        return self._populate_service_config(config_dict)

    def generate_config_from_file(
        self,
        service_config_file,
        name: Optional[str],
        description: Optional[str],
        healthcheck_url: Optional[str] = None,
    ) -> ServiceConfig:
        if not os.path.exists(service_config_file):
            raise click.ClickException(f"Config file {service_config_file} not found.")

        with open(service_config_file, "r") as f:
            config_dict = yaml.safe_load(f)

        service_config = self._populate_service_config(config_dict)
        if name:
            service_config.name = name

        if description:
            service_config.description = description

        if healthcheck_url:
            service_config.healthcheck_url = healthcheck_url

        return service_config

    def deploy_from_config(self, service_config: ServiceConfig):
        project_id = infer_project_id(
            service_config.project_id, self.anyscale_api_client, self.log
        )

        service_config.runtime_env = override_runtime_env_for_local_working_dir(
            service_config.runtime_env, self.log
        )

        config_object = ProductionJobConfig(
            entrypoint=service_config.entrypoint,
            runtime_env=service_config.runtime_env,
            build_id=service_config.build_id,
            compute_config_id=service_config.compute_config_id,
            max_retries=service_config.max_retries,
        )

        service = self.api_client.apply_service_api_v2_decorated_ha_jobs_apply_service_put(
            CreateProductionService(
                name=service_config.name
                or "cli-job-{}".format(datetime.now().isoformat()),
                description=service_config.description or "Service updated from CLI",
                project_id=project_id,
                config=config_object,
                healthcheck_url=service_config.healthcheck_url,
                access=service_config.access,
            )
        ).result

        maximum_uptime_minutes = self._get_maximum_uptime_minutes(service)
        self.log.info(
            f"Maximum uptime is {self._get_maximum_uptime_output(maximum_uptime_minutes)} for clusters launched by this service."
            f"{self._get_additional_log_if_maximum_uptime_enabled(maximum_uptime_minutes)}"
        )
        self.log.info(
            f"Service {service.id} has been deployed. Current state of service: {service.state.current_state}."
        )
        self.log.info(
            f"Query the status of the service with `anyscale service list --service-id {service.id}`."
        )
        self.log.info(
            f'View the service in the UI at {get_endpoint(f"/services/{service.id}")}.'
        )

    def _get_maximum_uptime_minutes(self, service: ProductionJob) -> Optional[int]:
        compute_config: ComputeTemplate = self.api_client.get_compute_template_api_v2_compute_templates_template_id_get(
            service.config.compute_config_id
        ).result
        return compute_config.config.maximum_uptime_minutes

    def _get_maximum_uptime_output(self, maximum_uptime_minutes: Optional[int]) -> str:
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return f"set to {maximum_uptime_minutes} minutes"
        return "disabled"

    def _get_additional_log_if_maximum_uptime_enabled(
        self, maximum_uptime_minutes: Optional[int]
    ) -> str:
        if maximum_uptime_minutes and maximum_uptime_minutes > 0:
            return " This may cause disruptions. To disable, update the compute config."
        return ""

    def _populate_service_config(self, config_dict: Dict[str, Any]) -> ServiceConfig:
        if "ANYSCALE_EXPERIMENTAL_WORKSPACE_ID" in os.environ:
            if "ANYSCALE_SESSION_ID" in os.environ:
                cluster = self.anyscale_api_client.get_cluster(
                    os.environ["ANYSCALE_SESSION_ID"]
                ).result
                # If the job configs are not specified, infer them from the workspace:
                if "build_id" not in config_dict and "cluster_env" not in config_dict:
                    config_dict["build_id"] = cluster.cluster_environment_build_id
                if "project_id" not in config_dict:
                    config_dict["project_id"] = cluster.project_id
                if (
                    "compute_config" not in config_dict
                    and "compute_config_id" not in config_dict
                ):
                    config_dict["compute_config_id"] = cluster.cluster_compute_id

        service_config = ServiceConfig.parse_obj(config_dict)
        return service_config

    def list(
        self,
        include_all_users: bool,
        include_archived: bool,
        name: Optional[str],
        service_id: Optional[str],
        project_id: Optional[str],
        max_items: int,
    ) -> None:
        self.job_controller.list(
            include_all_users,
            name,
            service_id,
            project_id,
            include_archived=include_archived,
            max_items=max_items,
            is_service=True,
        )

    def archive(self, service_id: Optional[str], service_name: Optional[str]) -> None:
        self.job_controller.archive(service_id, service_name, is_service=True)

    def terminate(self, service_id: Optional[str], service_name: Optional[str]) -> None:
        self.job_controller.terminate(service_id, service_name, is_service=True)
